#    get_recent_pdbe.py
#    Copyright (C) 2011  Bernhard Lohkamp
#    Copyright 2015, 2016 by Medical Research Council
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


# FIXME: review this initiallisations
import gobject
# some timeout setting (used?)
import socket
# FIXME, depening on what we download!?
socket.setdefaulttimeout(10)
#sys.setcheckinterval(10) # doesnt seem to make much of a difference...

import queue
max_queue = 14
coot_queue = queue.Queue()

import _thread
safe_print = _thread.allocate_lock()
def print_thread(*txt):
    safe_print.acquire()
    print(" ".join(map(str, txt)))
    safe_print.release()

global coot_pdbe_image_cache_dir
coot_pdbe_image_cache_dir = "coot-pdbe-images"  # can be shared (dir should
                                                # be writable by sharers)

def coot_pdbe():
    pass

#
def get_recent_json(file_name):

    import json  # maybe should be globally imported
    
    if not os.path.isfile(file_name):
        print("file not found", file_name)
        return False
    else:
        fin = open(file_name, 'r')
        s = json.load(fin)
        fin.close()
        return s

# geometry is an improper list of ints.
# 
# return the h_box of the buttons.
#
# a button is a list of [label, callback-thunk, text-description]
#
# If check-button-label is False, don't make one, otherwise create with with
# the given label and "on" state.
#
def dialog_box_of_buttons_with_async_ligands(window_name, geometry,
                                             buttons, close_button_label):

    """geometry is an improper list of ints.
    
    return the h_box of the buttons.
    
    a button is a list of [label, callback-thunk, text-description]
    
    If check-button-label is False, don't make one, otherwise create with with
    the given label and "on" state.
    """

    # the async function is evaluated here
    #
    def add_button_info_to_box_of_buttons_vbox_for_ligand_images(button_info, vbox):
        button_label = button_info[0]
        callback = button_info[1]
        active_button_label_func = button_info[2] \
                                   if (len(button_info) > 2) \
                                   else False  # it doesnt have one
        button = gtk.Button()
        button_hbox         = gtk.HBox(False, 0)   # pixmaps and labels get added here.
        protein_ribbon_hbox = gtk.HBox(False, 0)  # do we need a box here?!
        ligands_hbox        = gtk.HBox(False, 0)

        button_hbox.append(gtk.Label(button_label))
        # or just button.set_label?!
        button_hbox.append(ligands_hbox)
        button_hbox.append(protein_ribbon_hbox)
        button.add(button_hbox)
        button.connect("clicked", callback)
        # this is for function, but thread here!?!
        #if callable(active_button_label_func):
        #   active_button_label_func(button_hbox,
        #                             ligands_hbox,
        #                             protein_ribbon_hbox)
        # 
        # if active_button_label_func:
        # ignore nmr for now
        if not callable(active_button_label_func):
            entry_id = active_button_label_func[0]
            ligand_tlc_list = active_button_label_func[1]
            make_active_ligand_button(entry_id, ligand_tlc_list,
                                      ligands_hbox, protein_ribbon_hbox)
            # we pass the button to update in thread!
        
        vbox.append(button)
        button.show()

        # the 'new' version, we threaded download and add the
    # images to the button's hboxes
    def make_active_ligand_button(entry_id, ligand_tlc_list,
                                  ligands_hbox, protein_ribbon_hbox):

        global coot_pdbe_image_cache_dir
        # main line
        image_size = 100
        for tlc in ligand_tlc_list:
            # image_url = "http://www.ebi.ac.uk/pdbe-srv/pdbechem/image/showNew?code=" + \
            #             tlc + "&size=" + str(image_size)
            image_url = "https://www.ebi.ac.uk/pdbe/static/chem-files/" + \
                        tlc + "-" + str(image_size) + ".gif"
            image_name = os.path.join(coot_pdbe_image_cache_dir,
                                      (tlc + "-" + str(image_size) + ".gif"))
            cache_or_net_get_image(image_url, image_name, ligands_hbox)

        # now do the protein icon:
        image_size = 120
        # image_name_stub = entry_id + "_cbc" + str(image_size) + ".png"
        image_name_stub = "_deposited_chain_front_image-200x200.png"
        # image_url = "http://www.ebi.ac.uk/pdbe/entry-images/" + \
        #            image_name_stub
        image_url = "https://www.ebi.ac.uk/pdbe/static/entry/" + \
                    entry_id + image_name_stub
        
        entry_image_file_name = os.path.join(coot_pdbe_image_cache_dir,
                                             entry_id + image_name_stub)
        cache_or_net_get_image(image_url, entry_image_file_name,
                               protein_ribbon_hbox)
        

    # main line
    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    scrolled_win = gtk.ScrolledWindow()
    outside_vbox = gtk.VBox(False, 2)
    h_sep = gtk.HSeparator()
    inside_vbox = gtk.VBox(False, 0)

    window.set_default_size(geometry[0], geometry[1])
    window.set_title(window_name)

    inside_vbox.set_border_width(2)
    window.add(outside_vbox)

    outside_vbox.append(scrolled_win)
    scrolled_win.add_with_viewport(inside_vbox)
    scrolled_win.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_ALWAYS)
    if buttons:
        list(map(lambda button_info:
            coot_gui.add_button_info_to_box_of_buttons_vbox_for_ligand_images(button_info,
                                                                     inside_vbox),
            buttons))

    outside_vbox.set_border_width(2)
    outside_vbox.append(h_sep)
    ok_button = gtk.Button(close_button_label)
    outside_vbox.pack_end(ok_button, False, False, 0)

    def delete_event(*args):
        # global setting of inside vbox (locked?)?
        inside_vbox = False
        # shall empty the queue
        if not coot_queue.empty():
            while not coot_queue.empty():
                coot_queue.get()
                coot_queue.task_done()
        window.destroy()
        return False
    
    ok_button.connect("clicked", delete_event)
    window.connect("destroy", delete_event)
    window.show_all()
    # make a few downloaders, 14 seems to be a good number according to Paule
    for i in range(14):
        downloader_thread(i)
    return inside_vbox

def is_png_or_gif_qm(file_name):

    import imghdr
    if not os.path.isfile(file_name):
        return False
    else:
        return imghdr.what(file_name)

# Get image-name (caller doesn't care how) and when it is in place run func.
# This is a generally useful function, so it has been promoted outside of dig-table.
# mMMMmmm not sure if I want to run a lambda function here?! Let's see
# maybe this should be better queued, otherwise we may start lots of
# threads
#
def cache_or_net_get_image(image_url, image_name, hbox):

    import _thread  # try to get away with the simple ones
    import urllib.request, urllib.parse, urllib.error  # move to a proper place FIXME
    import threading

    def show_image_when_ready(image_name, hbox):
        if os.path.isfile(image_name):
            #apply(func)
            # instead of function we put image into vbox here
            pixmap = gtk.Image()
            if is_png_or_gif_qm(image_name):
               pixmap.set_from_file(image_name)
               hbox.append(pixmap)
               pixmap.show()
            return False  # stop
        else:
            return True # continue

    def get_image(url, file_name):
        try:
            file_name, url_info = urllib.request.urlretrieve(url, file_name)
        except socket.timeout:
            print_thread("BL ERROR:: timout download", url[-50:])
        except:
            print_thread("BL ERROR:: failed download", url[-50:])
    # this takes time!! Use queue instead!
    # thread.start_new_thread(get_image, (image_url, image_name))
    coot_queue.put((image_url, image_name))
    gobject.timeout_add(1000, show_image_when_ready, image_name, hbox)

def downloader_thread(thread_no):
    
    import urllib.request, urllib.parse, urllib.error

    def get_image():
        while True:
            try:
                url, file_name = coot_queue.get()
                try:
                    file_name, url_info = urllib.request.urlretrieve(url, file_name)
                except socket.timeout:
                    print_thread("BL ERROR:: timout download", url[-50:])
                except:
                    print_thread("BL ERROR:: failed download", url[-50:])
                coot_queue.task_done()
            except queue.Empty:
                # nothing left to do, die baby
                return
    _thread.start_new_thread(get_image, ())
    

# return refmac_result or False
#
def refmac_calc_sfs_make_mtz_with_columns(pdb_in_file_name, mtz_file_name,
                                          mtz_refmaced_file_name,
                                          f_col, sigf_col, r_free_col):

    global refmac_extra_params
    
    refmac_stub = os.path.join("coot-refmac",
                               coot_utils.strip_path(coot_utils.file_name_sans_extension(pdb_in_file_name)))
    pdb_out_file_name = refmac_stub + "-refmaced.pdb"
    mtz_out_file_name = mtz_refmaced_file_name
    extra_cif_lib_filename = ""
    imol_refmac_count = 0
    swap_map_colours_post_refmac = 0
    imol_mtz_molecule = False
    show_diff_map_flag = 0
    phase_combine_flag = 0
    phib_fom_pair = False
    force_n_cycles = 0
    make_molecules_flag = 0 # ??

    save_refmac_extra_params = refmac_extra_params
    if isinstance(refmac_extra_params, list):
        refmac_extra_params.append("MAKE NEWLIGAND CONTINUE")
    else:
        refmac_extra_params = ["MAKE NEWLIGAND CONTINUE"]
        
    refmac_result = refmac.run_refmac_by_filename(pdb_in_file_name,
                                           pdb_out_file_name,
                                           mtz_file_name, mtz_out_file_name,
                                           extra_cif_lib_filename,
                                           imol_refmac_count,
                                           swap_map_colours_post_refmac,
                                           imol_mtz_molecule,
                                           show_diff_map_flag,
                                           phase_combine_flag,
                                           phib_fom_pair,
                                           force_n_cycles,
                                           make_molecules_flag,
                                           "", f_col, sigf_col, r_free_col)

    # restore refmac-extra-params to what it used to be
    #
    refmac_extra_params = save_refmac_extra_params

    if os.path.isfile(mtz_refmaced_file_name):
        return refmac_result
    else:
        return False

def refmac_calc_sfs_make_mtz(pdb_in_file_name, mtz_file_name,
                             mtz_refmaced_file_name):

    return refmac_calc_sfs_make_mtz_with_columns(pdb_in_file_name, mtz_file_name,
                                                 mtz_refmaced_file_name,
                                                 "F.F_sigF.F", "F.F_sigF.sigF",
                                                 "Rfree.Flag.flag")

# include_get_sfs_flag is either "no-sfs" or "include-sfs"
#
def pdbe_get_pdb_and_sfs_cif(include_get_sfs_flag,
                             entry_id, method_string=""):

    import urllib.request, urllib.parse, urllib.error
    import time

    global download_thread_status
    download_thread_status = None     # start

    # At the moment PDBe wants lower case entry ids?! So we make them
    entry_id = entry_id.lower()

    coot_download_dir = coot_utils.get_directory("coot-download")
    if (not coot_download_dir):
        coot.info_dialog("Failed to make download directory")
    else:
        # do it!

        # just a small bit of abstraction.
        #
        def make_and_draw_map_local(refmac_out_mtz_file_name):
            coot.make_and_draw_map(refmac_out_mtz_file_name, "FWT", "PHWT", "", 0, 0)
            coot.make_and_draw_map(refmac_out_mtz_file_name, "DELFWT", "PHDELWT", "", 0, 1)

        def get_sfs_run_refmac(sfs_cif_url, sfs_cif_file_name,
                               sfs_mtz_file_name, pdb_file_name,
                               refmac_out_mtz_file_name,
                               cif_progress_bar, window):

            global download_thread_status

            #
            def convert_to_mtz_and_refmac(sfs_cif_file_name,
                                          sfs_mtz_file_name,
                                          pdb_file_name):

                global download_thread_status
                # OK, let's run convert to mtz and run refmac
                #
                download_thread_status = "converting-to-mtz"
                convert_status = coot.mmcif_sfs_to_mtz(sfs_cif_file_name,
                                                  sfs_mtz_file_name)
                if not (convert_status == 1):
                    # why cant we make a dialog?! (if this is threaded)
                    txt = "WARNING:: Failed to convert " +  \
                          sfs_cif_file_name + " to an mtz file"
                    print(txt)
                    #info_dialog(txt)
                    download_thread_status = "fail"
                else:
                    # first check cancel
                    if not (download_thread_status == "cancelled"):
                        # why extra function!?
                        #refmac_inner(pdb_file_name, sfs_mtz_file_name,
                        #             refmac_out_mtz_file_name)
                        download_thread_status = "running-refmac-for-phases"
                        refmac_result = refmac_calc_sfs_make_mtz(pdb_file_name,
                                                                 sfs_mtz_file_name,
                                                                 refmac_out_mtz_file_name)
                        print("      refmac-result: ", refmac_result)

                        # if refmac_result is good? (is tuple not list)
                        # good enough if it's not false?!
                        if not (isinstance(refmac_result, tuple)):
                            download_thread_status = "fail-refmac"
                        else:
                            # make map
                            # cant do here!! we are in thread!!
                            #make_and_draw_map_local(refmac_out_mtz_file_name)
                            download_thread_status = "done"  #??

            # main line get_sfs_run_refmac
            print("in get_sfs_run_refmac", sfs_cif_file_name, end=' ')
            sfs_mtz_file_name, pdb_file_name, refmac_out_mtz_file_name

            # check for cached results: only run refmac if
            # output file does not exist or is empty
            if (os.path.isfile(refmac_out_mtz_file_name) and
                os.stat(refmac_out_mtz_file_name).st_size > 0):
                # using cached result, i.e. skip
                download_thread_status = "done"
            else:
                # the coot refmac interface writes its
                # output in coot-refmac directory.  If
                # that doesn't exist and we can't make
                # it, then give up.
                #
                if (not coot.make_directory_maybe("coot_refmac") == 0):
                    coot.info_dialog("Can't make output directory coot-refmac")
                else:
                    if (os.path.isfile(sfs_cif_file_name) and
                        os.stat(sfs_cif_file_name).st_size > 0):
                        # we have sfs_cif_file_name already
                        download_thread_status = "done-download" # instruct to?
                        # needed
                    else:
                        # need to get sfs_mtz_file_name
                        download_thread_status = "downloading-sfs"
                        cif_thread = _thread.start_new_thread(download_file_and_update_widget,
                                                             (sfs_cif_url,
                                                              sfs_cif_file_name,
                                                              cif_progress_bar,
                                                              window))
                        def check_cif_download_thread():
                           if download_thread_status == "downloading-sfs":
                              return True
                           else:
                              # done
                              if (download_thread_status == "done-download"):
                                 convert_to_mtz_and_refmac(sfs_cif_file_name,
                                                           sfs_mtz_file_name,
                                                           pdb_file_name)

                                 return False
                              if (download_thread_status == "cancelled"):
                                 return False
                              else:
                                 return True
                        gobject.idle_add(check_cif_download_thread)

            
        # return a list of the progress bars and the window 
        # 
        # (the pdb-file-name and sfs-cif-file-name are passed so
        # that the cancel button knows what transfers to cancel (if
        # needed)).
        # FIXME:: all needed?!?
        def progress_dialog(pdb_file_name, sfs_cif_file_name):
            window = gtk.Window(gtk.WINDOW_TOPLEVEL)
            dialog_name = "Download and make SFS for " + entry_id
            main_vbox = gtk.VBox(False, 6)
            cancel_button = gtk.Button("  Cancel  ")
            buttons_hbox = gtk.HBox(False, 2)
            pdb_hbox = gtk.HBox(False, 6)
            cif_hbox = gtk.HBox(False, 6)
            refmac_hbox = gtk.HBox(False, 6)
            pdb_label = gtk.Label("Download PDB:  ")
            cif_label = gtk.Label("Download SFs cif:")
            refmac_label = gtk.Label("Running Refmac:")
            # FIXME: shouldnt we rather change the label than making
            # or adding a new one?!
            refmac_fail_label = gtk.Label("Running Refmac Failed")
            fail_label = gtk.Label("Something Went Wrong")
            pdb_progress_bar = gtk.ProgressBar()
            cif_progress_bar = gtk.ProgressBar()
            refmac_progress_bar = gtk.ProgressBar()

            pdb_execute_icon = gtk.image_new_from_stock("gtk-execute", 1)
            cif_execute_icon = gtk.image_new_from_stock("gtk-execute", 1)
            refmac_execute_icon = gtk.image_new_from_stock("gtk-execute", 1)
            pdb_good_icon = gtk.image_new_from_stock("gtk-ok", 1)
            cif_good_icon = gtk.image_new_from_stock("gtk-ok", 1)
            refmac_good_icon = gtk.image_new_from_stock("gtk-ok", 1)

            pdb_fail_icon = gtk.image_new_from_stock("gtk-no", 1)
            cif_fail_icon = gtk.image_new_from_stock("gtk-no", 1)
            refmac_fail_icon = gtk.image_new_from_stock("gtk-no", 1)
            h_sep = gtk.HSeparator()

            window.set_title(dialog_name)
            buttons_hbox.append(cancel_button)
            pdb_hbox.append(pdb_label)
            pdb_hbox.append(pdb_progress_bar)
            # NOTE: shouldnt we rather replace the icons?! Not sure
            pdb_hbox.append(pdb_execute_icon)
            pdb_hbox.append(pdb_good_icon)
            pdb_hbox.append(pdb_fail_icon)
            cif_hbox.append(cif_label)
            cif_hbox.append(cif_progress_bar)
            cif_hbox.append(cif_execute_icon)
            cif_hbox.append(cif_good_icon)
            cif_hbox.append(cif_fail_icon)
            refmac_hbox.append(refmac_label)
            refmac_hbox.append(refmac_progress_bar)
            refmac_hbox.append(refmac_execute_icon)
            refmac_hbox.append(refmac_good_icon)
            refmac_hbox.append(refmac_fail_icon)

            main_vbox.append(pdb_hbox)
            main_vbox.append(cif_hbox)
            main_vbox.append(refmac_hbox)
            main_vbox.append(refmac_fail_label)
            main_vbox.append(fail_label)
            main_vbox.append(h_sep)
            main_vbox.append(buttons_hbox)
            main_vbox.set_border_width(6)

            window.add(main_vbox)
            window.set_border_width(4)
            window.show_all()
            # hide the icons
            pdb_good_icon.hide()
            cif_good_icon.hide()
            refmac_good_icon.hide()
            pdb_fail_icon.hide()
            cif_fail_icon.hide()
            refmac_fail_icon.hide()
            refmac_fail_label.hide()
            fail_label.hide()

            pdb_execute_icon.set_sensitive(False)
            cif_execute_icon.set_sensitive(False)
            refmac_execute_icon.set_sensitive(False)

            def cancel_button_cb(*args):
                global download_thread_status
                window.destroy()
                download_thread_status = "cancelled"

            cancel_button.connect("clicked", cancel_button_cb)

            # return a dictionary (cant we do a lookup directly?!)
            return {"window": window,
                    "pdb_progress_bar": pdb_progress_bar,
                    "cif_progress_bar": cif_progress_bar,
                    "refmac_progress_bar": refmac_progress_bar,
                    "refmac_fail_label": refmac_fail_label,
                    "fail_label": fail_label,
                    "pdb_execute_icon": pdb_execute_icon,
                    "cif_execute_icon": cif_execute_icon,
                    "refmac_execute_icon": refmac_execute_icon,
                    "pdb_fail_icon": pdb_fail_icon,
                    "cif_fail_icon": cif_fail_icon,
                    "refmac_fail_icon": refmac_fail_icon,
                    "pdb_good_icon": pdb_good_icon,
                    "cif_good_icon": cif_good_icon,
                    "refmac_good_icon": refmac_good_icon}

        # main line
        coords_type = ".ent"  # can/will be ".cif"
#       pdb_url = "http://www.ebi.ac.uk/pdbe-srv/view/files/" + \
#                       entry_id + ".ent"
#       sfs_cif_url = "http://www.ebi.ac.uk/pdbe-srv/view/files/r" + \
#                       entry_id + "sf.ent"
        pdb_url     = "https://www.ebi.ac.uk/pdbe/entry-files/pdb" + \
                      entry_id + coords_type
        sfs_cif_url = "https://www.ebi.ac.uk/pdbe/entry-files/r"   + entry_id + "sf.ent"
        pdb_file_name = os.path.join("coot-download", entry_id + coords_type)
        sfs_cif_file_name = os.path.join("coot-download", "r" + entry_id + "sf.cif")
        sfs_mtz_file_name = os.path.join("coot-download", "r" + entry_id + "sf.mtz")
        refmac_out_mtz_file_name = os.path.join("coot-download",
                                                "r" + entry_id + "-refmac.mtz")
        refmac_log_file_name = "refmac-from-coot-" + \
                               str(refmac_count) + ".log" # set in refmac.run_refmac_by_filename
        progr_widgets = progress_dialog(pdb_file_name, sfs_cif_file_name)
        window = progr_widgets["window"]
        pdb_progress_bar = progr_widgets["pdb_progress_bar"]
        cif_progress_bar = progr_widgets["cif_progress_bar"]
        refmac_progress_bar = progr_widgets["refmac_progress_bar"]
        refmac_fail_label = progr_widgets["refmac_fail_label"]
        fail_label = progr_widgets["fail_label"]
        pdb_execute_icon = progr_widgets["pdb_execute_icon"]
        cif_execute_icon = progr_widgets["cif_execute_icon"]
        refmac_execute_icon = progr_widgets["refmac_execute_icon"]
        pdb_good_icon = progr_widgets["pdb_good_icon"]
        cif_good_icon = progr_widgets["cif_good_icon"]
        refmac_good_icon = progr_widgets["refmac_good_icon"]
        pdb_fail_icon = progr_widgets["pdb_fail_icon"]
        cif_fail_icon = progr_widgets["cif_fail_icon"]
        refmac_fail_icon = progr_widgets["refmac_fail_icon"]

        def download_file_and_update_widget(url, file_name,
                                            progress_bar, window):
            global download_thread_status
            def update_progressbar_in_download(numblocks, blocksize,
                                               filesize, progress_bar):
                global download_thread_status
                percent = min((numblocks*blocksize*100)/filesize, 100)
                #print "BL DEBUG:: update pg bar" , percent, numblocks, blocksize, filesize, progress_bar
                if download_thread_status == "cancelled":
                    sys.exit()
                if (percent > 100):
                    percent = 100
                gobject.idle_add(progress_bar.set_fraction, percent/100.)
                # in the end should change icons?!

            def sleeper():
                global download_thread_status
                time.sleep(0.01)
                if ("downloading" in download_thread_status):
                    return True # continue downloading
                else:
                    return False
            try:
                #print "BL DEBUG:: start download", url
                gobject.idle_add(sleeper)
                file_name_local, url_info = urllib.request.urlretrieve(url, file_name,
                                                               lambda nb, bs, fs, progress_bar=progress_bar:
                                                               update_progressbar_in_download(nb, bs, fs, progress_bar))
                download_thread_status = "done-download"
            except socket.timeout:
                print("BL ERROR:: timout download", url[-50:])
                download_thread_status = "fail"
            except IOError:
                print("BL ERROR:: ioerror downloading", url[-50:])
                download_thread_status = "fail"
            except:
                if (not download_thread_status == "cancelled"):
                    print("BL ERROR:: general problem downloading", url[-50:])
                    download_thread_status = "fail"
                else:
                    print("BL INFO:: cancelled download")

        # or shall this be in the timeout function!?
        def update_refmac_progress_bar(refmac_progress_bar, log_file_name):
            # refmac puts out 100 lines of text before it starts running. 
            # Let's not count those as progress of the computation (otherwise 
            # we jump to 22% after a fraction of a second).
            max_lines = 350     # thats 450 - 100
            n_lines = coot_utils.file_n_lines(log_file_name)
            if n_lines:  # check for number?
                n_lines_rest = n_lines - 100.
                if (n_lines > 0):
                    f = n_lines_rest / max_lines
                    if (f <= 1.0 and
                        f >= 0.):
                        refmac_progress_bar.set_fraction(f)

        def update_dialog_and_check_download(*args):
            global download_thread_status

            if (download_thread_status == "downloading-pdb"):
                pdb_execute_icon.set_sensitive(True)

            if (download_thread_status == "downloading-sfs"):
                pdb_progress_bar.set_fraction(1.)
                pdb_execute_icon.hide()
                pdb_good_icon.show()
                cif_execute_icon.set_sensitive(True)

            if (download_thread_status == "converting-to-mtz"):
                pdb_progress_bar.set_fraction(1.)
                cif_progress_bar.set_fraction(1.)

            if (download_thread_status == "running-refmac-for-phases"):
                pdb_progress_bar.set_fraction(1.)
                cif_progress_bar.set_fraction(1.)
                refmac_execute_icon.set_sensitive(True)
                cif_execute_icon.hide()
                cif_good_icon.show()
                # de-sensitise cancel?!
                update_refmac_progress_bar(refmac_progress_bar,
                                           refmac_log_file_name)

            if (download_thread_status == "fail-refmac"):
                refmac_fail_label.show()
                refmac_execute_icon.hide()
                refmac_fail_icon.show()
                return False  # dont continue

            # generic fail (don't turn off execute icons because we 
            # don't know *what* failed. (Not so useful).
            if (download_thread_status == "fail"):
                fail_label.show()
                return False # dont continue

            if (download_thread_status == "done"):
                make_and_draw_map_local(refmac_out_mtz_file_name)
                window.destroy()
                return False  # we are done!
            return True

        # could update progressbar here too!?
        gobject.timeout_add(200, update_dialog_and_check_download,
                            window,
                            pdb_execute_icon, pdb_good_icon, pdb_fail_icon,
                            cif_execute_icon, cif_good_icon, cif_fail_icon,
                            pdb_progress_bar, cif_progress_bar,
                            refmac_progress_bar, refmac_log_file_name,
                            refmac_fail_label, fail_label)

        # Try with individual threads just for downloading, alternative use
        # Paul's approach and make functions to be executed in the the idle
        # function

        def start_cif_download():
             global download_thread_status
             # read the pdb
             imol = coot.read_pdb(pdb_file_name)
             if not coot_utils.valid_model_molecule_qm(imol):
                 s = "Oops - failed to correctly read " + pdb_file_name
                 coot.info_dialog(s)
                 download_thread_status = "fail"
             if (include_get_sfs_flag != "include-sfs"):
                 # an NMR structure
                 #
                 # FIXME:: better later as timeout_add
                 # coot.read_pdb(pdb_file_name)
                 download_thread_status = "done"  #?
                 #print "BL DEBUG:: NMR structure!?"
             else:
                 # An X-ray structure
                 #
                 # now read the sfs cif and if that is good then
                 # convert to mtz and run refmac, but for now, let's
                 #  just show the PDB file.
                 if not (os.path.isfile(sfs_cif_file_name) and
                         os.path.isfile(refmac_out_mtz_file_name)):
                     # download and run refmac
                     _thread.start_new_thread(get_sfs_run_refmac, (sfs_cif_url, sfs_cif_file_name,
                                                                  sfs_mtz_file_name, pdb_file_name,
                                                                  refmac_out_mtz_file_name,
                                                                  cif_progress_bar, window))
                 else:
                     # OK, files are here already.
                     #
                     # FIXME:: again for later
                     # coot.read_pdb(pdb_file_name) # already done above
                     # coot.auto_read_make_and_draw_maps(refmac_out_mtz_file_name) # done in the updater.
                     download_thread_status = "done"

        def check_pdb_download_thread():
          if download_thread_status == "downloading-pdb":
              return True
          else:
              # done
              if (download_thread_status == "done-download"):
                  # really done, now start the cif download
                  start_cif_download()
                  return False
              else:
                  if (download_thread_status == "cancelled"):
                      return False
                  return True
        # Get the PDB file if we don't have it already.
        if os.path.isfile(pdb_file_name):
            download_thread_status = "done-download"
        else:
            download_thread_status = "downloading-pdb"
            pdb_thread = _thread.start_new_thread(download_file_and_update_widget,
                                                         (pdb_url, pdb_file_name, pdb_progress_bar,
                                                          window))
        gobject.idle_add(check_pdb_download_thread)
            #print "BL DEBUG:: done with pdb thread?!"

    
def recent_structure_browser(t):

    import string
    import _thread
    global refmac_count
    global download_thread_status
    download_thread_status = None  # initiallise

    def get_dic_all_entries(js):
        # simple minded, we know where the entries are,
        # so return the dictionary of these
        # print "extract something from ", js
        return js["grouped"]["pdb_id"]["groups"]

    n_atoms_limit_small_ligands = 6
    
    # truncate (if needed) and newlinify string
    #
    def pad_and_truncate_name(name):
        sl = len(name)
        if (sl < 60):
            return "\n     " + name
        else:
            return "\n     " + name[0:60] + "..."
        
    # as above but no preceeding newline and tab
    def truncate_name(name):
        sl = len(name)
        if (sl < 70):
            return name
        else:
            return name[0:70] + "..."

    #
    def make_ligands_string(ligand_string_list):

        if not ligand_string_list:
            return ""
        else:
            ligand_string = ""
            for ligand in ligand_string_list:
                ligand_string = ligand_string + "   " + ligand + "\n"
            if len(ligand_string) > 0:
                return "\nLigands:\n" + ligand_string
            else:
                return ""

    # The idea here is that we have a list of ligand tlcs.  From that,
    # we make a function, which, when passed a button-hbox (you can put
    # things into a button) will download images from pdbe and add them
    # to the button.  We download the images in the background
    # (sub-thread) and put them in coot-pdbe-ligand-images directory.
    # When/if the tranfer of the file is completed and good, set the
    # status to a function.
    #
    # a timeout function watches the value of the status and when it is
    # a function, it runs it.  (The function puts the image in the
    # button-hbox.)
    # 
    # Also we construct the url of a ribbon picture give the entry id
    #
    def make_active_ligand_button_func(entry_id, ligand_tlc_list):
        # do we need this?!
        return


    # dic is a dictionary with keys: "Title", "Resolution", "EntryID", etc.
    #
    def handle_pdb_entry_entity(dic_wrapper):

        # return a string. Return "" on failure
        def make_authors_string(author_list):
            
            citation = ""
            if not author_list:
                citation = "No Authors listed"
            else:
                citation = coot_utils.string_append_with_spaces(author_list)
            return citation

        global coot_pdbe_image_cache_dir
        coot.make_directory_maybe(coot_pdbe_image_cache_dir)

        # now make a button list (a label and what to do)
        # entry_id = str(dic["EntryID"])
        groupValue = dic_wrapper["groupValue"]
        if not isinstance(groupValue, str):
            print("failed to get groupValue")
        else:
            dic = dic_wrapper["doclist"]["docs"][0]
            entry_id = str(dic["pdb_id"])

            entry_label = entry_id if entry_id else "missing entry id"  # should not happen, the latter!?
            method_item = dic["experimental_method"]
            try:
                resolution_item = dic["resolution"]
            except KeyError:
                resolution_item = False
            title_item = dic["title"]
            try: 
                authors_string = make_authors_string(dic["pubmed_author_list"])
            except KeyError:
                authors_string = ""
            # method_label = method_item if method_item else "Unknown method"
            # should not happen, the latter!?
            method_label = method_item[0]
            title_label = truncate_name(title_item) if title_item else ""
            authors_label = authors_string
            try: 
                ligands = dic["compound_id"]
            except KeyError:
                ligands = []
            ligands_string = make_ligands_string(ligands)
            ligand_tlc_list = ligands
            resolution_string = ("     Resolution: " + str(resolution_item)) \
                                if resolution_item else ""

            if False:
                print("   title_label:", title_label)
                print("   entry_id:", entry_id)
                print("   method_label:", method_label)
                print("   resolution_string:", resolution_string)
                print("   authors_label:", authors_label)
                print("   ligands_string:", ligands_string)
            label = entry_id + "\n" + title_label + "\n" + method_label + \
                    resolution_string + "\n" + authors_label + ligands_string

            if (method_label == "X-ray diffraction"):
                return[label,
                       lambda func: pdbe_get_pdb_and_sfs_cif("include-sfs",
                                                             entry_id,
                                                             method_label),
                       # we dont do a funtion here, but use the args
                       # when we construct the button!?
                       [entry_id, ligand_tlc_list]]
            else:
                return[label,
                       lambda func: pdbe_get_pdb_and_sfs_cif("no-sfs",
                                                             entry_id,
                                                             method_label),
                       [entry_id, ligand_tlc_list]]
                    # this one (may) has to be lambda func too
                    #lambda func2:
                    #make_active_ligand_button_func(entry_id, ligand_tlc_list)]

    # return a list of buttons
    #
    def handle_pdb_entry_entities(dic):
        return list(map(handle_pdb_entry_entity, dic))
                    
    # main line!?
    aa = get_dic_all_entries(t)
    button_list = handle_pdb_entry_entities(aa)
    coot_gui.dialog_box_of_buttons_with_async_ligands("Recent Entries", [700, 500],
                                             button_list, " Close ")
    
def recent_entries_progress_dialog():

    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    dialog_name = "Getting recent entries list..."
    main_vbox = gtk.VBox(False, 4)
    label = gtk.Label("Getting recent entries list...")
    progress_bar = gtk.ProgressBar()

    window.set_title(dialog_name)
    main_vbox.append(label)
    main_vbox.append(progress_bar)
    window.add(main_vbox)
    window.set_border_width(4)
    window.show_all()

    return (progress_bar, window)
    

# Sameer Velankar says to get this file for the latest releases
# "http://www.ebi.ac.uk/pdbe-apps/jsonizer/latest/released/" (note the end "/")
#
def pdbe_latest_releases_gui():

    import threading
    import urllib.request, urllib.parse, urllib.error


    url = "https://www.ebi.ac.uk/pdbe/search/latest/select?facet=true&q=*%3A*&group=true&group.field=pdb_id&group.ngroups=true&&json.nl=map&fq=document_type%3Alatest_pdb&fq=entry_type:%28new%20OR%20revised%29&wt=json&fl=pdb_id,release_date,resolution,number_of_bound_molecules,experimental_method,citation_title,citation_doi,pubmed_author_list,journal,title,entry_type&rows=-1"
    url = "https://www.ebi.ac.uk/pdbe/search/latest/select?facet=true&q=*%3A*&group=true&group.field=pdb_id&group.ngroups=true&&json.nl=map&fq=document_type%3Alatest_pdb&fq=entry_type:%28new%20OR%20revised%29&wt=json&fl=pdb_id,compound_id,release_date,resolution,number_of_bound_molecules,experimental_method,citation_title,citation_doi,pubmed_author_list,journal,title,entry_type&rows=-1"

    # url = "http://www.ebi.ac.uk/pdbe-apps/jsonizer/latest/released/"
    json_file_name = "latest-releases.json"

    coot.add_status_bar_text("Retrieving list of latest releases...")
    # FIXME:: progress bar!?
    progress_bars = recent_entries_progress_dialog()

    class MyURLopener(urllib.request.FancyURLopener):
        def http_error_default(self, url, fp, errcode, errmsg, headers):
            # handle errors the way you'd like to
            # raise StandardError, ("File not found?")
            pass
    
    class GetUrlThread(threading.Thread):
        def __init__(self, url, file_name, progressbar, window):
            self.url = url
            self.file_name = file_name
            self.window = window
            self.progressbar = progressbar
            self.status = None
            threading.Thread.__init__(self)

        def get_url_status(self):
            return self.status

        def update_function(self, count, blockSize, totalSize):
            # total size is -1! Baeh!!
            percent = int(count*blockSize*100/totalSize)
            gobject.idle_add(self.update_progressbar, percent)
        
        def update_progressbar(self, count):
            self.progressbar.pulse()
            return False

        def destroy_window(self):
            self.window.destroy()
            
        def run(self):
            try:
                self.file_name = coot_urlretrieve(self.url, self.file_name, self.update_function)
                self.status = 0 #?
            except socket.timeout:
                print("BL ERROR:: timout download", self.url)
            except IOError:
                print("BL ERROR:: ioerror")
            except:
                self.status = 1
                # FIXME here dies with the thread. need to go to main thread
                gobject.idle_add(info_dialog, "Failed to get recent entry list from server.")
            gobject.idle_add(self.destroy_window)

    thread = GetUrlThread(url, json_file_name,
                          progress_bars[0], progress_bars[1])
    _thread.start()

    def start_table():
        if _thread.get_url_status() == 0:
            recent_structure_browser(get_recent_json(json_file_name))
            return False  # stop
        return True  # continue

    def run_sleeper():
        if _thread.get_url_status() == 0:
            return False   # stop
        time.sleep(0.02)
        return True # continue
    
    gobject.idle_add(start_table)
    # most likely not needed
#    gobject.timeout_add(50, run_sleeper)





#    what_check.py
#    Copyright (C) 2008  Bernhard Lohkamp, The University of York
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


# "  53" -> "53", " " -> ""
def strip_leading_spaces(str):
    return str.lstrip()

# consider for coot_utils
def go_to_residue_by_spec(imol, spec):

    def atom_spec_qm(spec):
        return (len(spec) == 5)

    # main line
    coot_utils.set_go_to_atom_molecule(imol)
    if (atom_spec_qm(spec)):
        atom_name = spec[3]
    else:
        atom_name = " CA "

    coot_utils.set_go_to_atom_chain_residue_atom_name(spec[0],
                                           spec[1],
                                           atom_name)

# consider for coot_utils
def residue_spec_to_string(spec):
    print("DEBUG:: in residue_spec_to_string() with spec", spec)
    if len(spec) == 4:
        return (spec[1] + str(spec[2]) + spec[3])
    else:
        return (spec[0] + str(spec[1]) + spec[2])

# make this internal 
# 
# where line is something like  "  53 HIS  (  53 ) A12"
#                                   ignore   xxxxI Cyy
# xxxx is 4char resno, I insertion code, C chain id, yy is model number
# 
# return a residue spec, or False
# 
def get_flip_residue(line):
    l = len(line)
    if (l < 19):
        return False
    else:
        resno_string = line[11:15]
        try:
            resno = int(resno_string)
        except:
            return False
        inscode_pre = line[15:16]
        if (inscode_pre == " "):
            inscode = ""
        else:
            inscode = inscode_pre
        chain_id = line[18:19]
        print("found", resno, inscode, chain_id)
        return [chain_id, resno, inscode]

def convert_to_button_info(problemed_res_list):
    #print "convert_to_button_info got", problemed_res_list
    button_label      = problemed_res_list[1]
    problem_string    = problemed_res_list[0]
    action_func       = problemed_res_list[2]
    residue_spec_list = problemed_res_list[3]

    o = map(lambda x: [problem_string, button_label, action_func, x],
            residue_spec_list)
    #print "convert_to_button_info returns", o
    return o

# A problemed-res-list is improper pair where the car is a pair of
# string, one describing the problem, and the other being the "fix
# it" button label.  The cdr is a list of residue specs.
# 
# A problemed_flip_list_list is a list of those things.
# 
def problem_residues2dialog(imol, problemed_res_list_list):
    if (problemed_res_list_list):
        def button_list():
            ret = []
            converted_list = map(convert_to_button_info, problemed_res_list_list)
            for problemed_list in converted_list:
                for flip_res_info in problemed_list:
                    flip_res_spec  = flip_res_info[3]
                    button_label   = flip_res_info[1]
                    action_func    = flip_res_info[2]
                    problem_string = flip_res_info[0]
                    ret.append([[problem_string + " " + residue_spec_to_string(flip_res_spec),
                                 "go_to_residue_by_spec(" + str(imol) + ", " + str(flip_res_spec) + ")"],
                                [button_label,
                                 str(action_func) + "(" + str(imol) + ", " + str(flip_res_spec) + ")"]])

            #print "BL DEBUG:: returning button_list", ret
            return ret

        coot_gui.dialog_box_of_pairs_of_buttons(imol,
                                       "What-Check Report",
                                       [270, 300],
                                       # buttons is a list of: [[button_1_label, button_1_action,],
                                       #                        [button_2_label, button_2_action]]
                                       # The button_1_action function is a string.
                                       # The button_2_action function is a string.
                                       button_list(),
                                       "  Close  ")
        
# action is either 'gui' (then we get a gui) or 'apply-actions', then the model
# modifications are automatically applied.
#
def parse_check_db(imol, file_name, action):

    #
    global my_flip, my_rsr, my_delete_atom, my_rotamer_search
    def my_flip(imol_local, res_spec):
        coot.do_180_degree_side_chain_flip(imol_local,
                                      res_spec[0],
                                      res_spec[1],
                                      res_spec[2],
                                      "")

    def my_rsr(imol_local, res_spec):
        coot_utils.with_auto_accept([refine_zone, imol_local, res_spec[0], res_spec[1],
                           res_spec[1], ""])

    #
    def my_delete_atom(imol, atom_spec):
        ls = [imol] + atom_spec
        print("calling delete_atom with args", ls)
        coot.delete_atom(*ls)

    #
    def my_rotamer_search(imol, res_spec):
        chain_id = res_spec[0]
        resno    = res_spec[1]
        ins_code = res_spec[2]
        altloc   = ""

        imol_map = coot.imol_refinement_map()
        if (coot_utils.valid_map_molecule_qm(imol_map)):
            print( "rotamer search with imol: %s chain: %s resno: %s inscode: %s" \
                  %(imol, chain_id, resno, ins_code))
            coot.auto_fit_best_rotamer(resno, altloc, ins_code, chain_id, imol, imol_map, 1, 0.01)
        else:
            coot.add_status_bar_text("No refinement map available, no rotamer fit")

    # return False or a residue spec:
    def line2residue_spec(line):
        if (len(line) >21):
            resno_string = line[26:30]
            try:
                resno = int(resno_string)
            except:
                print("BL WARNING:: no int resno -> return")
                return False
            inscode_pre = line[40:41]
            if (inscode_pre == "_"):
                inscode = ""
            else:
                inscode = inscode_pre
            chain_id_pre = line[19:20]
            chain_id = " " if (chain_id_pre == "_") else chain_id_pre
            return [chain_id, resno, inscode]

    # return False or an atom_spec
    # an atom_spec is [chain_id, resno, ins_code, atom_name, alt_conf]
    #
    def line2atom_spec(line):
        if (len(line) > 53):
            resno_string  = line[26:30]
            try:
                resno = int(resno_string)
            except:
                return False
            inscode_pre = line[40:41]
            inscode = "" if (inscode_pre == "_") else inscode_pre
            chain_id_pre = line[19:20]
            chain_id = "" if (chain_id_pre == "_")  else chain_id_pre
            atom_name_pre = line[47:51]
            atom_name = atom_name_pre
            alt_conf = ""
            print("found atom name %s resno %s (from %s) inscode %s (%s) chain_id %s" \
                  %(atom_name, resno, resno_string, inscode_pre, inscode, chain_id))
            return [chain_id, resno, inscode, atom_name, alt_conf]

    def apply_actions(imol, action_list):

        for action in action_list:
            func = action[0]
            baddie_list = action[1]
            print( "BL DEBUG:: action and baddie_list", action, baddie_list)
            imol_list = map(lambda x: imol, baddie_list)
            print( "BL DEBUG:: imol_list", imol_list)
            print( "applying %s to %s" %(func, baddie_list))
            print( "BL DEBUG:: len of lists", len(imol_list), len(baddie_list))
            map(func, imol_list, baddie_list)

    # main line
    print( "parse_whatcheck_report: parsing output of what-if...")
    found_flip_table     = False
    found_h2ohbo_table   = False
    found_BMPCHK_table   = False
    found_NAMCHK_table   = False
    found_PLNCHK_table   = False
    h2o_h_bond_less_list = []
    plane_list           = []
    flip_list            = []
    namchk_list          = []
    bump_list            = []

    if (os.path.isfile(file_name)):
        fin = open(file_name, 'r')
        lines = fin.readlines()
        fin.close()
        for line in lines:
            line = line.rstrip('\n')  # remove trailing \n
            if (line == "//"):
                found_flip_table     = False
                found_h2ohbo_table   = False
                found_BMPCHK_table   = False
                found_NAMCHK_table   = False
                found_PLNCHK_table   = False

            if (found_flip_table):
                if (len(line) > 12):
                    attr_type = line[0:11]
                    if (attr_type == "   Name   :"):
                        #print "found flip name", line
                        res = line2residue_spec(line)
                        if (res):
                            flip_list.append(res)
            
            if (line == "CheckID   : HNQCHK"):
                found_flip_table = True

            if (found_h2ohbo_table):
                if (len(line) > 12):
                    attr_type = line[0:11]
                    if (attr_type == "   Name   :"):
                        #print "found h2obmo name", line
                        atom = line2atom_spec(line)
                        if (atom):
                            h2o_h_bond_less_list.append(atom)
            
            if (line == "CheckID   : H2OHBO"):
                found_h2ohbo_table = True

            if (found_BMPCHK_table):
                if (len(line) > 12):
                    attr_type = line[0:11]
                    if (attr_type == "   Name   :"):
                        #print "found bump", line
                        res = line2residue_spec(line)
                        if (res):
                            bump_list.append(res)
            
            if (line == "CheckID   : BMPCHK"):
                found_BMPCHK_table = True
            
            if (found_NAMCHK_table):
                if (len(line) > 12):
                    attr_type = line[0:11]
                    if (attr_type == "   Name   :"):
                        res = line2residue_spec(line)
                        if (res):
                            namchk_list.append(res)
            
            if (line == "CheckID   : NAMCHK"):
                found_NAMCHK_table = True
            
            if (found_PLNCHK_table):
                if (len(line) > 12):
                    attr_type = line[0:11]
                    if (attr_type == "   Name   :"):
                        res = line2residue_spec(line)
                        if (res):
                            plane_list.append(res)
            
            if (line == "CheckID   : PLNCHK"):
                found_PLNCHK_table = True
            
        #print "Here flip_list is", flip_list
        #print "Here h2o-h-bond-less-list is", h2o_h_bond_less_list
        #print "Here bump_list is", bump_list
        print( "Here plane_list is", plane_list)
        if (action == 'gui'):
            problem_residues2dialog(imol, [
                ["Needs flipping", "Flip it", "my_flip", flip_list],
                ["Water with no H-bonds", "Delete it", "my_delete_atom", h2o_h_bond_less_list],
                ["Has Bumps", "Rotamer Search", "my_rotamer_search", bump_list],
                ["Nomenclature error", "180 degree flip", "my_flip", namchk_list]])
        elif (action == 'apply-actions'):
            apply_actions(imol, [
                # [my_flip, flip_list],
                # [my_delete_atom, h2o_h_bond_less_list],
                # [my_rotamer_search, bump_list],
                [my_rsr, plane_list]
                # [my_flip, namchk_list]
                ])
        else:
            print( "INFO:: wrong action input")
                     
            
    else:
        print( "file not found", file_name)





# rename coot_utils.py functions

def residue_spec_to_chain_id(rs):
    if not isinstance(rs, list):
        return False
    else:
        if (len(rs) == 3):
            return rs[0]
        else:
            if (len(rs) == 4):
                return rs[1]
            else:
                return False

def residue_spec_to_res_no(rs):
    if not isinstance(rs, list):
        return False
    else:
        if (len(rs) == 3):
            return rs[1]
        else:
            if (len(rs) == 4):
                return rs[2]
        return False

def residue_spec_to_ins_code(rs):
    if not isinstance(rs, list):
        return False
    else:
        if (len(rs) == 3):
            return rs[2]
        else:
            if (len(rs) == 4):
                return rs[3]
        return False


def residue_atom_to_position(ra):
    if not isinstance(ra, list):
        return False
    else:
        return ra[2]




#!/usr/bin/env python
"""A class for storing data about a suite of RNA"""

# Copyright 2010, 2011 Kevin Keating
# 
# Licensed under the Educational Community License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
# http://www.osedu.org/licenses/ECL-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

class Suite:
    """A class for storing data about a suite of RNA
    
    NOTE:
        No actual data is stored in the Suite object other than a link to the starting and
        ending nucleotides.  All functions defined here simply call the appropriate
        Nucleotide object function.
    """
    
    
    def __init__(self, startingNuc, endingNuc):
        """Initialize a Suite object
        
        ARGUMENTS:
            startingNuc - a Nucleotide object representing the starting nucleotide of the suite
            endingNuc   - a Nucleotide object representing the ending nucleotide of the suite
        RETURNS:
            an initialized Suite object
        """
        
        self.startingNuc = startingNuc
        self.endingNuc   = endingNuc
    
    def sugarDist(self):
        """Calculate the sugar distance for the suite"""
        return self.endingNuc.startingSugarDist()
    
    def theta(self):
        """Calculate theta for the suite"""
        return self.startingNuc.theta()
    
    def eta(self):
        """Calculate eta for the suite"""
        return self.endingNuc.eta()
    
    def startingPperp(self):
        """Calculate the starting base-phosphate perpendicular (P-perp) distance for the suite"""
        return self.startingNuc.pperp()
        
    def endingPperp(self):
        """Calculate the ending base-phosphate perpendicular (P-perp) distance for the suite"""
        return self.endingNuc.pperp()
    
    def startingPhosDist(self):
        """Calculate the starting phosphate-phosphate distance for the suite"""
        return self.startingNuc.phosDist()
        
    def endingPhosDist(self):
        """Calculate the ending phosphate-phosphate distance for the suite"""
        return self.endingNuc.phosDist()
    
    def suiteNum(self):
        """The suite number"""
        return self.endingNuc.resNum




def hole_ify():
    global start_pos, end_pos
    start_pos = False
    end_pos = False

    def status_bar_pos(position, pos_type):
        s = "Hole %s point set: (%6.2f %6.2f %6.2f)" %(pos_type,
                                                       position[0],
                                                       position[1],
                                                       position[2])
        coot.add_status_bar_text(s)

    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    vbox = gtk.VBox(False, 0)
    hbox = gtk.VBox(False, 0)
    hbox_pos_buttons = gtk.HBox(False, 0)
    hbox_calc_cancel_buttons = gtk.HBox(False, 0)
    start_button = gtk.Button("Set Start Point")
    end_button   = gtk.Button("Set End Point")
    calculate_button = gtk.Button("Calculate")
    cancel_button = gtk.Button("Cancel")
    option_menu_and_model_mol_list = coot_gui.generic_molecule_chooser(hbox,
                                                              "HOLE-ify molecule: ")

    window.add(vbox)
    hbox_pos_buttons.pack_start(start_button, True, True, 6)
    hbox_pos_buttons.pack_start(end_button, True, True, 6)
    hbox_calc_cancel_buttons.pack_start(calculate_button, True, True, 6)
    hbox_calc_cancel_buttons.pack_start(cancel_button, True, True, 6)
    vbox.pack_start(hbox, True, True, 6)
    vbox.pack_start(hbox_pos_buttons, True, True, 6)
    vbox.pack_start(hbox_calc_cancel_buttons, True, True, 6)

    def start_button_cb(*args):
        global start_pos
        start_pos = coot_utils.rotation_centre()
        print( "Start pos set to:", start_pos)x
        status_bar_pos(start_pos, "start")
        
    start_button.connect("clicked", start_button_cb)

    def end_button_cb(*args):
        global end_pos
        end_pos = coot_utils.rotation_centre()
        print "End pos set to:", end_pos
        status_bar_pos(end_pos, "end")
        
    end_button.connect("clicked", end_button_cb)

    def delete_event(*args):
        window.destroy()
        return False
    
    def calculate_button_cb(*args):
        global start_pos, end_pos
        imol = coot_gui.get_option_menu_active_molecule(*option_menu_and_model_mol_list)
        if isinstance(imol, int):
            print( start_pos, end_pos)
            if not isinstance(start_pos, list):
                coot.add_status_bar_text("Start position not set")
            else:
                if not isinstance(end_pos, list):
                    coot.add_status_bar_text("End position not set")
                else:
                    # main?
                    print( "hole", imol, start_pos, end_pos)
                    colour_map_multiplier = 1
                    colour_map_offset = 0
                    hole_args = [imol] + start_pos + end_pos + \
                                [colour_map_multiplier, colour_map_offset, 1, 1 ]
                    coot.hole(*hole_args)
                    delete_event()
                    
    calculate_button.connect("clicked", calculate_button_cb)

    cancel_button.connect("clicked", delete_event)

    window.show_all()

if (have_coot_python):
    if coot_python.main_menubar():
        menu = coot_gui.coot_menubar_menu("Test Hole")
        coot_gui.add_simple_coot_menu_menuitem(
            menu,
            "Test",
            lambda func: hole_ify()
            )



#!/usr/bin/env python
"""Create a select map dialog box, similar to Coot's show_select_map_dialog(), but allow for a callback."""

# Copyright 2011 Kevin Keating
# 
# Licensed under the Educational Community License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
# http://www.osedu.org/licenses/ECL-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

from coot import molecule_name, graphics_n_molecules, is_valid_map_molecule, imol_refinement_map, set_imol_refinement_map
import gtk


def selectMapDialog(callback = None):
    """Display a GUI for selecting the refinement map.
    
    OPTIONAL ARGUMENTS:
        callback - a function to call when the user clicks OK
    RETURNS:
        None
    """
    
    #create a list of map molecules and their names
    mapList = []
    for imol in xrange(graphics_n_molecules()):
        if is_valid_map_molecule(imol):
            mapList.append(imol)
    
    #if we have no maps, then display an error and don't invoke the callback
    if len(mapList) == 0:
        warningMessage = gtk.MessageDialog(type=gtk.MESSAGE_ERROR, buttons=gtk.BUTTONS_OK, message_format="Please load a map before using RCrane.")
        warningMessage.set_title("RCrane Error")
        warningMessage.run()
        warningMessage.destroy()
    
    else:
        #if we have one or more maps
        dialog = gtk.Dialog(title="Select Map for Fitting", buttons=(gtk.STOCK_OK, gtk.RESPONSE_OK))
        
        #create a combo box that lists maps
        mapListStore = gtk.ListStore(int, str)
        for imol in mapList:
            mapListStore.append([imol, str(imol) + " " + molecule_name(imol)])
        mapCombo = gtk.ComboBox(mapListStore)
        mapRender = gtk.CellRendererText()
        mapCombo.pack_start(mapRender, True)
        mapCombo.add_attribute(mapRender, 'text', 1)
        
        #select the current refinemnt map, or the first map if no map has been selected
        currentMap = imol_refinement_map()
        if currentMap == -1:
            set_imol_refinement_map(mapListStore[0][0])
            mapCombo.set_active(0)
        else:
            #figure out which combobox row corresponds to the currently selected map
            comboIndex = [i[0] for i in mapListStore].index(currentMap)
            mapCombo.set_active(comboIndex)
        
        #connect the changed signal for the combobox
        mapCombo.connect("changed", __changeMap)
        
        #add the label and frame
        chooseLabel = gtk.Label("Choose a Map:")
        frame = gtk.Frame()
        frameBox = gtk.VBox(False, 0)
        frame.add(frameBox)
        frameBox.pack_start(chooseLabel, False, False, 2)
        frameBox.pack_start(mapCombo, False, False, 2)
        
        #pack things into the top area of the dialog box
        dialog.vbox.pack_start(frame, False, False, 2)
        dialog.vbox.show_all()
        
        #connect the callback function
        if callback is not None:
            #if there is a callback function and the user clicks OK, then call the callback
            #if the user clicks the close button, then don't call the callback
            def checkResponse(dialog, responseID, callback):
                if responseID == gtk.RESPONSE_OK: callback()
            dialog.connect("response", checkResponse, callback)
        
        #create the dialog box
        dialog.run()
        dialog.destroy()
        

def __changeMap(mapCombo):
    """Respond to the user selecting a map
        
    ARGUMENTS:
        mapCombo - the ComboBox object for map selection
    RETURNS:
        None
    EFFECTS:
        sets the refinement map to the newly selected map
    """
    
    #figure out the newly selected map
    mapListStore  = mapCombo.get_model()
    selectedMapIter = mapCombo.get_active_iter()
    selectedImol = mapListStore.get_value(selectedMapIter, 0)
    
    #set the refinement map
    set_imol_refinement_map(selectedImol)



# fitting.py
# Copyright 2005, 2006, 2007 by Bernhard Lohkamp 
# Copyright 2004, 2005, 2006, 2008, 2009 by The University of York
# Copyright 2009 by Bernhard Lohkamp
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import coot
import coot_utils

# import coot_gui - coot_gui - this file is for scripting
# import coot_toolbuttons # 20220910-PE FIXME later

# For each residue in the protein (molecule number @var{imol}), do a
# rotamer fit and real-space refinement.  Update the graphics and
# rotate the scene at each residue for eye candy goodness.
#
# Note that residue with alt confs do not undergo auto-fit-rotamer.
# This is because that autofit-rotamer then refine will tend to put
# both rotamers into the same place.  Not good.  It seems a
# reasonable expectation that residues with an alternate conformation
# are already resonably well-fitted.  So residues with alternate
# conformations undergo only real space refinement.
#
# This is simple-minded and outdated now we have the interruptible
# version (below).
#
def fit_protein(imol, rotamer_only=False, animate=True):

    coot_utils.set_go_to_atom_molecule(imol)
    coot.make_backup(imol) # do a backup first
    backup_mode = coot.backup_state(imol)
    imol_map  = coot.imol_refinement_map()
    replacement_state = coot.refinement_immediate_replacement_state()

    if imol_map == -1:
        coot.info_dialog("Oops.  Must set a map to fit")
    else:

        coot.turn_off_backup(imol)
        coot.set_refinement_immediate_replacement(1)
          
        for chain_id in coot_utils.chain_ids(imol):
         if (not coot_utils.is_solvent_chain_qm(imol,chain_id)):
             n_residues = coot.chain_n_residues(chain_id,imol)
             print("There are %(a)i residues in chain %(b)s" % {"a":n_residues,"b":chain_id})

             for serial_number in range(n_residues):
                  
                res_name = coot.resname_from_serial_number(imol, chain_id, serial_number)
                res_no = coot.seqnum_from_serial_number(imol, chain_id, serial_number)
                ins_code = coot.insertion_code_from_serial_number(imol, chain_id, serial_number)
                res_atoms = coot.residue_info_py(imol, chain_id, res_no, ins_code)
                
                if (ins_code is not None):
                    if (len(res_atoms) > 3):
                    #if (not res_name=="HOH"): not needed if only refining 3 or  more atoms
                        for alt_conf in coot_utils.residue_alt_confs(imol, chain_id, res_no, ins_code):
                            print("centering on ",chain_id,res_no," CA")
                            coot_utils.set_go_to_atom_chain_residue_atom_name(chain_id,res_no,"CA")
                            if animate:
                                coot.rotate_y_scene(30, 0.3) # n-frames frame-interval(degrees)
                            if (alt_conf == ""):
                                coot.auto_fit_best_rotamer(res_no, alt_conf, ins_code, chain_id, imol,
                                                      imol_map, 1, 0.1)
                            if (imol_map >= 0 and
                                not rotamer_only):
                                coot.refine_zone(imol, chain_id, res_no, res_no, alt_conf)
                                coot.accept_regularizement()
                            if animate:
                                coot.rotate_y_scene(30, 0.3)
      
    if (replacement_state == 0):
        coot.set_refinement_immediate_replacement(0)
    if (backup_mode == 1):
        coot.turn_on_backup(imol)

# Paul: 20090517: thinking about making the fit-protein function
# interruptible with a toolbar button press.  How do we do that? 
# fit_protein needs to be split into 2 parts, one, that generates a
# list of residues specs the other that does a refinement given a
# residue spec, then we run and idle function that calls
# fit_residue_by_spec or each spec in turn and at the end return
#

# These 2 variables are used by multi-refine function(s), called by
# idle functions to refine just one residue.
# 
global continue_multi_refine
global multi_refine_spec_list
global multi_refine_idle_proc
global multi_refine_stop_button
global multi_refine_cancel_button
global multi_refine_continue_button
global multi_refine_separator
continue_multi_refine = False
multi_refine_spec_list = []
multi_refine_idle_proc = False
multi_refine_stop_button = False
multi_refine_cancel_button = False
multi_refine_continue_button = False
multi_refine_separator = False

# Return a list of residue specs
#
# chain-specifier can be a string, where it is the chain of interest.
# or 'all-chains, where all chains are chosen.
#
def fit_protein_make_specs(imol, chain_specifier):
    specs_list = []
    if coot.is_valid_model_molecule(imol):
        pass
    chain_list = []
    if chain_specifier == 'all-chains':
        chain_list = coot_utils.chain_ids(imol)
    else:
        chain_list = [chain_specifier]
    for ch_id in chain_list:
        serial_number_list = list(range(coot.chain_n_residues(ch_id, imol)))
        for serial_number in serial_number_list:
            res_no   = coot.seqnum_from_serial_number(imol, ch_id, serial_number)
            ins_code = coot.insertion_code_from_serial_number(imol, ch_id, serial_number)
            if ins_code is not None:
                specs_list.append([imol, ch_id, res_no, ins_code])
        
    return specs_list

    
def fit_protein_make_specs_from_residue_range(imol, chain_id,
                                              res_no_start, res_no_end):
    if coot.valid_model_molecule_qm(imol):
        if res_no_end > res_no_start:
            ret = [[chain_id, res_no, ""] for res_no in range(res_no_start, res_no_end + 1 )]
            return ret
    return []
    
    
def fit_protein_fit_function(res_spec, imol_map):

    imol     = res_spec[0]
    chain_id = res_spec[1]
    res_no   = res_spec[2]
    ins_code = res_spec[3]

    res_name = coot.residue_name(imol, chain_id, res_no, ins_code)
    if isinstance(res_name, str):
        if (res_name != "HOH"):
            for alt_conf in coot_utils.residue_alt_confs(imol, chain_id, res_no, ins_code):
                print("centering on", chain_id, res_no, "CA")
                coot_utils.set_go_to_atom_chain_residue_atom_name(chain_id, res_no, "CA")
                coot.rotate_y_scene(10, 0.3) # n_frames frame_interval(degrees)        
                res_atoms = coot.residue_info_py(imol, chain_id, res_no, ins_code)
                if (len(res_atoms) > 3):
                    # if (not res_name == "HOH"):
                    # not needed as we only refine more than 3 atom res
                    if (alt_conf == ""):
                        with coot_utils.NoBackups(imol):
                            coot.auto_fit_best_rotamer(res_no, alt_conf, ins_code, chain_id, imol,
                                                  imol_map, 1, 0.1)
                    if (coot_utils.valid_map_molecule_qm(imol_map)):
                        with coot_utils.NoBackups(imol):
                            with AutoAccept():
                                coot.refine_zone(imol, chain_id, res_no, res_no, alt_conf)
                    coot.rotate_y_scene(10, 0.3)


def fit_protein_stepped_refine_function(res_spec, imol_map, use_rama = False):

    imol     = res_spec[0]
    chain_id = res_spec[1]
    res_no   = res_spec[2]
    ins_code = res_spec[3]
    current_rama_state = coot.refine_ramachandran_angles_state()
    
    for alt_conf in coot_utils.residue_alt_confs(imol, chain_id, res_no, ins_code):
        if use_rama:
            coot.set_refine_ramachandran_angles(1)
        res_name = coot.residue_name(imol, chain_id, res_no, ins_code)
        if (not res_name == "HOH"):
            print("centering on", chain_id, res_no, "CA")
            coot_utils.set_go_to_atom_chain_residue_atom_name_full(chain_id, res_no,
                                                        ins_code, "CA",
                                                        alt_conf)
            coot.rotate_y_scene(10, 0.3) # n_frames frame_interval(degrees)
            with coot_utils.NoBackups(imol):
                with AutoAccept():
                    coot.refine_auto_range(imol, chain_id, res_no, alt_conf)
            coot.rotate_y_scene(10, 0.3)    

    coot.set_refine_ramachandran_angles(current_rama_state)

    
def fit_protein_rama_fit_function(res_spec, imol_map):
    # BL says: make it more generic
    fit_protein_stepped_refine_function(res_spec, imol_map, True)

# func is a refinement function that takes 2 args, one a residue
# spec, the other the coot.imol_refinement_map.  e.g. fit_protein_fit_function
#
def interruptible_fit_protein(imol, func):

    # import gobject
    global multi_refine_spec_list
    global continue_multi_refine
    global multi_refine_idle_proc
    global multi_refine_stop_button
    global multi_refine_separator
    specs = fit_protein_make_specs(imol, 'all-chains')
    if specs:
        # lets make a backup before we start
        coot.make_backup(imol)

        if False:
           # 20230501-PE old stuff commented out on gtk3 branch merge
           # multi_refine_separator = coot_toolbuttons.add_coot_toolbar_separator()
           # multi_refine_stop_button = coot_gui.coot_toolbar_button("Stop", "stop_interruptible_fit_protein()", "gtk-stop")
           pass

        multi_refine_spec_list = specs
        def idle_func():
            global multi_refine_spec_list
            global continue_multi_refine
            global multi_refine_idle_proc
            global multi_refine_stop_button
            global multi_refine_separator
            if not multi_refine_spec_list:
                #set_visible_toolbar_multi_refine_stop_button(0)
                #set_visible_toolbar_multi_refine_continue_button(0)
                multi_refine_stop_button.destroy()
                multi_refine_stop_button = False
                multi_refine_separator.destroy()
                multi_refine_separator = False
                return False
            if continue_multi_refine:
                imol_map = coot.imol_refinement_map()
                func(multi_refine_spec_list[0], imol_map)
                del multi_refine_spec_list[0]
                return True
            else:
                # finish what we have been doing first before we stop
                coot.accept_regularizement()
                return False
        # gobject.idle_add(idle_func)
        multi_refine_idle_proc = idle_func
        coot.set_go_to_atom_molecule(imol)


# For each residue in chain chain-id of molecule number imol, do a
# rotamer fit and real space refinement of each residue.  Don't
# update the graphics while this is happening (which makes it faster
# than fit-protein, but much less interesting to look at).
#
def fit_chain(imol, chain_id):

    coot.make_backup(imol)
    backup_mode = coot.backup_state(imol)
    imol_map = coot.imol_refinement_map()
    alt_conf = ""
    replacement_state = coot.refinement_immediate_replacement_state()

    coot.turn_off_backup(imol)
    coot.set_refinement_immediate_replacement(1)

    if imol_map == -1:
       print("WARNING:: fit-chain undefined imol-map. Skipping!!")
    else:
       n_residues = coot.chain_n_residues(chain_id,imol)
       for serial_number in range(n_residues):
           res_name = coot.resname_from_serial_number(imol,chain_id,serial_number)
           res_no = coot.seqnum_from_serial_number(imol,chain_id,serial_number)
           ins_code = coot.insertion_code_from_serial_number(imol,chain_id,serial_number)
           if ins_code is not None:
               res_atoms = coot.residue_info_py(imol, chain_id, res_no, ins_code)
               if (len(res_atoms) > 3):  # actually then we dont need the water check any more?!
                   #if (not res_name == "HOH"):
                       print("centering on ", chain_id, res_no, " CA")
                       coot.set_go_to_atom_chain_residue_atom_name(chain_id,res_no,"CA")
                       coot.auto_fit_best_rotamer(res_no, alt_conf, ins_code,
                                                  chain_id, imol, imol_map, 1, 0.1)
                       if (imol_map >= 1):
                           coot.refine_zone(imol,chain_id,res_no,res_no,alt_conf)
                           coot.accept_moving_atoms()

    if replacement_state == 0:
        coot.set_refinement_immediate_replacement(0)
    if backup_mode == 1:
        coot.turn_on_backup(imol)

# As fit_chain, but only for a residue range from resno_start to resno_end
def fit_residue_range(imol, chain_id, resno_start, resno_end):

    coot.make_backup(imol)
    backup_mode = coot.backup_state(imol)
    imol_map = coot.imol_refinement_map()
    alt_conf = ""
    replacement_state = coot.refinement_immediate_replacement_state()

    coot.turn_off_backup(imol)
    coot.set_refinement_immediate_replacement(1)
    
    if (imol_map == -1):
       print("WARNING:: fit-chain undefined imol-map. Skipping!!")
    else:
       n_residues = coot.chain_n_residues(chain_id,imol)
       ins_code = ""
       for res_no in coot_utils.number_list(resno_start, resno_end):
           print("centering on ", chain_id, res_no, " CA")
           coot.set_go_to_atom_chain_residue_atom_name(chain_id, res_no, "CA")
           res_atoms = coot.residue_info_py(imol, chain_id, res_no, ins_code)
           if (len(res_atoms) > 3):
               coot.auto_fit_best_rotamer(res_no, alt_conf, ins_code, chain_id,
                                          imol, imol_map, 1, 0.1)
               if (imol_map >= 1):
                   coot.refine_zone(imol,chain_id,res_no,res_no,alt_conf)
                   coot.accept_regularizement()

    if (replacement_state == 0):
        coot.set_refinement_immediate_replacement(0)
    if (backup_mode == 1):
          coot.turn_on_backup(imol)


# For each residue in the solvent chains of molecule number
# @var{imol}, do a rigid body fit of the water to the density.
#
# BL says: we pass *args where args[0]=imol and args[1]=animate_qm (if there)
def fit_waters(imol, animate_qm = False):

    print("animate?:", animate_qm)
    imol_map = coot.imol_refinement_map()
    do_animate_qm = False
    if (animate_qm):
        do_animate_qm = True

    print("do_animate?: ", do_animate_qm)
 
    if (imol_map != -1):
        replacement_state = coot.refinement_immediate_replacement_state()
        backup_mode = coot.backup_state(imol)
        alt_conf = ""

        coot.turn_off_backup(imol)
        coot.set_refinement_immediate_replacement(1)
        coot.set_go_to_atom_molecule(imol)

        # refine waters
        for chain_id in coot_utils.chain_ids(imol):
            if (coot_utils.is_solvent_chain_qm(imol, chain_id)):
                n_residues = coot.chain_n_residues(chain_id, imol)
                print("There are %(a)i residues in chain %(b)s" % {"a":n_residues,"b":chain_id})
                for serial_number in range(n_residues):
                    res_no = coot.seqnum_from_serial_number(imol,chain_id,serial_number)
                    if do_animate_qm:
                        res_info = coot.residue_info_py(imol, chain_id, res_no, "")
                        if not res_info == []:
                            atom = res_info[0]
                            atom_name = atom[0]
                            coot.set_go_to_atom_chain_residue_atom_name(chain_id, res_no, atom_name)
                            coot.refine_zone(imol, chain_id, res_no, res_no, alt_conf)
                            coot.rotate_y_scene(30, 0.6)	# n-frames frame-interval(degrees)
                    else:
                        coot.refine_zone(imol,chain_id,res_no,res_no,alt_conf)
                    coot.accept_regularizement()

        if (replacement_state == 0):
            coot.set_refinement_immediate_replacement(0)
        if (backup_mode == 1):
            coot.turn_on_backup(imol)

    else:
        coot.add_status_bar_text("You need to define a map to fit the waters")


# BL thingy:
# as fit_waters, but will only fit a range of waters (start to end).
# speeds things up when refining a lot of waters
#
def fit_waters_range(imol, chain_id, start, end):

    imol_map = coot.imol_refinement_map()
    if (imol_map != -1):
       replacement_state = coot.refinement_immediate_replacement_state()
       backup_mode = coot.backup_state(imol)
       alt_conf = ""

       coot.turn_off_backup(imol)
       coot.set_refinement_immediate_replacement(1)

       if (coot_utils.is_solvent_chain_qm(imol,chain_id)):
          serial_number = start
          while serial_number <= end:
                res_no = serial_number
                coot.refine_zone(imol,chain_id,res_no,res_no,alt_conf)
                coot.accept_regularizement()
                serial_number += 1

       if (replacement_state == 0):
          coot.set_refinement_immediate_replacement(0)
       if (backup_mode == 1):
          coot.turn_on_backup(imol)

# Step through the residues of molecule number imol and at each step
# do a residue range refinement (unlike fit-protein for example,
# which does real-space refinement for every residue).
#
# The step is set internally to 2.
#
def stepped_refine_protein(imol, res_step = 2):

    import types
    from types import IntType

    imol_map = coot.imol_refinement_map()
    if (not coot_utils.valid_map_molecule_qm(imol_map)):            
        coot.info_dialog("Oops, must set map to refine to")
    else:
        def refine_func(chain_id, res_no):
            #print "centering on ",chain_id,res_no," CA"
            coot.set_go_to_atom_chain_residue_atom_name(chain_id, res_no, "CA")
            coot.rotate_y_scene(30, 0.3) # n-frames frame-interval(degrees)
            coot.refine_auto_range(imol, chain_id, res_no, "")
            coot.accept_regularizement()
            coot.rotate_y_scene(30,0.3)
        stepped_refine_protein_with_refine_func(imol, refine_func, res_step)


# refine each residue with ramachandran restraints
#
def stepped_refine_protein_for_rama(imol):

    imol_map = coot.imol_refinement_map()
    if (not coot_utils.valid_map_molecule_qm(imol_map)):
        coot.info_dialog("Oops, must set map to refine to")
    else:
        current_rama_state  = coot.refine_ramachandran_angles_state()
        def refine_func(chain_id, res_no):
            coot.set_go_to_atom_chain_residue_atom_name(chain_id, res_no, "CA")
            coot.refine_auto_range(imol, chain_id, res_no, "")
            coot.accept_regularizement()

        coot.set_refine_ramachandran_angles(1)
        stepped_refine_protein_with_refine_func(imol, refine_func, 1)
        coot.set_refine_ramachandran_angles(current_rama_state)

#
def stepped_refine_protein_with_refine_func(imol, refine_func, res_step):

    coot_utils.set_go_to_atom_molecule(imol)
    coot.make_backup(imol)
    backup_mode = coot.backup_state(imol)
    alt_conf = ""
    imol_map = coot.imol_refinement_map()
    replacement_state = coot.refinement_immediate_replacement_state()

    if (imol_map == -1):
        # actually shouldnt happen as we set check the map earlier...
        coot.add_status_bar_text("Oops.  Must set a map to fit")
    else:
        coot.turn_off_backup(imol)
        coot.set_refinement_immediate_replacement(1)
        res_step = int(res_step)
        if (res_step <= 1):
            coot.set_refine_auto_range_step(1)
            res_step = 1
        else:
            coot.set_refine_auto_range_step(int(res_step / 2))

        for chain_id in coot_utils.chain_ids(imol):
            n_residues = coot.chain_n_residues(chain_id,imol)
            print("There are %(a)i residues in chain %(b)s" % {"a":n_residues,"b":chain_id})
            
            for serial_number in range(0, n_residues, res_step):
                res_name = coot.resname_from_serial_number(imol, chain_id, serial_number)
                res_no = coot.seqnum_from_serial_number(imol, chain_id, serial_number)
                ins_code = coot.insertion_code_from_serial_number(imol, chain_id, serial_number)
                if ins_code is not None:
                    print("centering on ", chain_id, res_no, " CA")
                    refine_func(chain_id, res_no)
                
        if (replacement_state == 0):
            coot.set_refinement_immediate_replacement(0)
        if (backup_mode == 1):
            coot.turn_on_backup(imol)


# The gui that you see after ligand finding. 
# 
def post_ligand_fit_gui():

     def test(imol):
       if not coot.is_valid_model_molecule(imol):
          return False
       else:
          name = coot.molecule_name(imol)
          if "Fitted ligand" in name:
             list = [name]
             # BL says: I think there must be a cleverer way!
             for i in coot_utils.molecule_centre(imol):
                list.append(i)
             return list
          else:
             return False 
     molecules_matching_criteria(lambda imol: test(imol))

# test_func is a function given one argument (a molecule number) that
# returns either False if the condition is not satisfied or something
# else if it is.  And that "something else" can be a list like
# [label, x, y, z]
# or 
# ["Bad Chiral", 0, "A", 23, "", "CA", "A"]
# 
# It is used in the create a button label and "what to do when the
# button is pressed".
# 
def molecules_matching_criteria(test_func):

    try:
        import pygtk
        pygtk.require("2.0")
        import gtk, pango

        # first count the number of fitted ligands, and post this if is
        # is greater than 0.

        passed_molecules = []
        for molecule_numbers in coot_utils.molecule_number_list():

            if (test_func(molecule_numbers)):
               passed_molecules.append(molecule_numbers)

        if (len(passed_molecules) > 0):
        # ok proceed

           def delete_event(*args):
               window.destroy()
               return False

           def centre_on_mol(widget, imol, name):
               s = "Centred on " + name
               coot.add_status_bar_text(s)
               centre = coot_utils.molecule_centre(imol)
               coot.set_rotation_centre(*centre)

           window = gtk.Window(gtk.WINDOW_TOPLEVEL)
           scrolled_win = gtk.ScrolledWindow()
           outside_vbox = gtk.VBox(False,2)
           inside_vbox = gtk.VBox(False,0)

           window.set_default_size(200,140)
           window.set_title("Fitted Ligands")
           # inside_vbox.set_border_width(2)

           window.add(outside_vbox)
           outside_vbox.append(scrolled_win)

           scrolled_win.add_with_viewport(inside_vbox)
           scrolled_win.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)

           # (format #t "debug:: passed-molecules ~s~%" passed-molecules)
           # ((("Fitted ligand #9" 68.4 11.9 4.6) 21)
           #  (("Fitted ligand #8" 68.3 12.8 8.1) 20))

           for molecule_numbers in passed_molecules:
             imol = molecule_numbers
             name = coot.molecule_name(imol)
             button = gtk.Button(label=str(name))
             inside_vbox.append(button)
             button.connect("clicked", centre_on_mol, imol, name)

           outside_vbox.set_border_width(6)
           ok_button = gtk.Button(label="OK")
           outside_vbox.append(ok_button)
           ok_button.connect_object("clicked",delete_event,window)
           window.connect("delete_event", delete_event)

           window.show_all()

        else:
           # no matching molecules
           coot.add_status_bar_text("No matching molecules!")

    except:
        print("BL WARNING:: no pygtk2. This function doesnt work!")

# This totally ignores insertion codes.  A clever algorithm would
# need a re-write, I think.  Well, we'd have at this end a function
# that took a chain_id res_no_1 ins_code_1 res_no_2 ins_code_2 
# 
# And refine-zone would need to be re-written too, of course.  So
# let's save that for a rainy day (days... (weeks)).
# 
# BL says:: this seems to ignore backup!
def refine_active_residue_generic(side_residue_offset):

    active_atom = active_residue()

    if not active_atom:
       print("No active atom")
    else:
       imol       = active_atom[0]
       chain_id   = active_atom[1]
       res_no     = active_atom[2]
       ins_code   = active_atom[3]
       atom_name  = active_atom[4]
       alt_conf   = active_atom[5]
    
       print("active-atom:", active_atom)
       imol_map = coot.imol_refinement_map()
       replacement_state = coot.refinement_immediate_replacement_state()
       if imol_map == -1:
          coot.info_dialog("Oops.  Must Select Map to fit to!")
       else:
          coot.set_refinement_immediate_replacement(1)
          coot.refine_zone(imol, chain_id, res_no - side_residue_offset,
                                      res_no + side_residue_offset,
                                      alt_conf)
          coot.accept_regularizement()
       if replacement_state == 0:
          coot.set_refinement_immediate_replacement(0)

# Function for keybinding. Speaks for itself
def refine_active_residue():
    refine_active_residue_generic(0)

# And another one
def refine_active_residue_triple():
    refine_active_residue_generic(1)


# For just one (this) residue, side-residue-offset is 0.
# 
def manual_refine_residues(side_residue_offset):

    active_atom = coot.active_residue()

    if not active_atom:
       print("No active atom")
    else:
       imol       = active_atom[0]
       chain_id   = active_atom[1]
       res_no     = active_atom[2]
       ins_code   = active_atom[3]
       atom_name  = active_atom[4]
       alt_conf   = active_atom[5]

    imol_map = coot.coot.imol_refinement_map()

    if (imol_map == -1):
        coot.info_dialog("Oops.  Must Select Map to fit to!")
    else:
        coot.refine_zone(imol, chain_id,
                         res_no - side_residue_offset,
                         res_no + side_residue_offset,
                         alt_conf)

# generic spherical refinement (use_map) (or regularization, dont use map):
def sphere_refine_regularize_generic(use_map=True, radius=3, expand=False):
    # from types import ListType
    active_atom = coot.active_residue_py()
    if (not active_atom):
        coot.add_status_bar_text("No active residue")
    else:
        if (use_map and not coot_utils.valid_map_molecule_qm(coot.imol_refinement_map())):
            coot.show_select_map_dialog()
        else:
            imol      = active_atom[0]
            chain_id  = active_atom[1]
            res_no    = active_atom[2]
            ins_code  = active_atom[3]
            atom_name = active_atom[4]
            alt_conf  = active_atom[5]
            centred_residue = active_atom[1:4]
            other_residues = coot.residues_near_residue_py(imol, centred_residue, radius)
            coot_utils.all_residues = [centred_residue]
            coot_utils.all_residues += other_residues

            # extend?
            if expand:
                print("in sphere_refine_regularize_generic, all_residues is", all_residues, "using radius", radius)
                coot_utils.all_residues.sort()
                tmp_ls = coot_utils.all_residues[:]
                for res in tmp_ls:
                    before_res = res[:]
                    after_res = res[:]
                    before_res[1] = before_res[1]-1
                    after_res[1] = after_res[1]+1
                    if not before_res in coot_utils.all_residues:
                        coot_utils.all_residues.append(before_res)
                    if not after_res in coot_utils.all_residues:
                        coot_utils.all_residues.append(after_res)
                coot_utils.all_residues.sort()  # not needed

            print("imol: %s residues: %s" %(imol, coot_utils.all_residues))
            if use_map:
                # don't use 'soft-mode/hard-mode' at the moment
                # (not sure how to integrate weight change into dragged refinement)
                coot.refine_residues_with_modes_with_alt_conf_py(imol, coot_utils.all_residues, "",
                                                                 '#soft-mode/hard-mode', False, False)
            else:
                coot.regularize_residues_py(imol, coot_utils.all_residues)

                # Sphere refinement (around radius)
#
def sphere_refine(radius=4.5, expand=False):
    sphere_refine_regularize_generic(True, radius, expand)

def sphere_refine_plus(radius=4.5):
    sphere_refine(radius, True)

# Sphere regularization (as above)
def sphere_regularize(radius=4.5, expand=False):
    sphere_refine_regularize_generic(False, radius, expand)

def sphere_regularize_plus(radius=4.5):
    sphere_regularize(radius, True)


def refine_tandem_residues():
    active_atom = coot.closest_atom_simple_py() # active_atom returns the CA if it can
    if not active_atom:
       print("No active atom")
    else:
       imol       = active_atom[0]
       chain_id   = active_atom[1]
       res_no     = active_atom[2]
       ins_code   = active_atom[3]
       atom_name  = active_atom[4]
       alt_conf   = active_atom[5]
       specs = []
       for ires in range(res_no-3, res_no+4):
           try:
              # test if the residue exists by looking for a residue name
              rn = coot.residue_name(imol, chain_id, ires, ins_code)
              if len(rn) > 0:
                  specs.append([chain_id, ires, ins_code])
           except TypeError as e:
               pass # no need to tell us
       refine_residues(imol, specs)


# Pepflip the active residue - needs a key binding
#
def pepflip_active_residue():
    active_atom = coot.closest_atom_simple_py() # active_atom returns the CA if it can
    if not active_atom:
       print("No active atom")
    else:
       imol       = active_atom[0]

       ca = closest_atom_raw() # don't map to CA

       chain_id   = ca[1]
       res_no     = ca[2]
       ins_code   = ca[3]
       atom_name  = ca[4]
       alt_conf   = ca[5]

       if (atom_name == " N  "): # PDBv3 fixme
           res_no -= 1;
       coot.pepflip(imol, chain_id, res_no, ins_code, alt_conf)

    

# Another cool function that needs a key binding
#
def auto_fit_rotamer_active_residue():

    active_atom = active_residue()

    if not active_atom:
       print("No active atom")
    else:
       imol       = active_atom[0]
       chain_id   = active_atom[1]
       res_no     = active_atom[2]
       ins_code   = active_atom[3]
       atom_name  = active_atom[4]
       alt_conf   = active_atom[5]
   
       print("active-atom:", active_atom)
       imol_map = coot.imol_refinement_map()
       replacement_state = coot.refinement_immediate_replacement_state()
       if imol_map == -1:
          coot.info_dialog("Oops.  Must Select Map to fit to!")
       else:
          coot.auto_fit_best_rotamer(res_no, alt_conf, ins_code, chain_id, imol, imol_map, 1, 0.1)


# Backrub rotamers for chain. After alignment mutation we should run this.
#
def backrub_rotamers_for_chain(imol, ch_id):

    """Backrub rotamers for chain. After alignment mutation we should run this."""

    coot.set_rotamer_search_mode(2) # ROTAMERSEARCHLOWRES
    coot.make_backup(imol)

    with coot_utils.NoBackups(imol):
        n_times = 2
        imol_map = coot.imol_refinement_map()
        if coot_utils.valid_map_molecule_qm(imol_map):
            n_res = coot.chain_n_residues(ch_id, imol)
            for i_round in range(n_times):
                for serial_number in range(n_res):
                    res_name = coot.resname_from_serial_number(imol, ch_id, serial_number)
                    res_no = coot.seqnum_from_serial_number(imol, ch_id, serial_number)
                    ins_code = coot.insertion_code_from_serial_number(imol, ch_id, serial_number)
                    print("debug res_name", res_name)
                    print("debug res_no", res_no)
                    print("debug ins_code", ins_code)
                    print("debug ch_id", ch_id)
                    if isinstance(ins_code, str):   # valid residue check :-)
                        if not res_name == "HOH":
                            # coot.auto_fit_best_rotamer(res_no, "", ins_code, ch_id, imol, imol_map, 1, 0.1)
                            coot.auto_fit_best_rotamer(imol, ch_id, res_no, ins_code, "", imol_map, 1, 0.1)


# Restrain the atoms in imol (in give range selection) to
# corresponding atoms in imol_ref.
# 
# atom_sel_type is either 'all' 'main-chain' or 'ca'
#
def add_extra_restraints_to_other_molecule(imol, chain_id,
                                           resno_range_start, resno_range_end,
                                           atom_sel_type, imol_ref):

    for res_no in range(resno_range_start, resno_range_end):
        pass # guess should do seomething?!?!

def add_extra_start_pos_restraints(imol, residue_spec, esd):

    ri = coot.residue_info_py(imol,
                              res_spec_utils.residue_spec_to_chain_id(residue_spec),
                              res_spec_utils.residue_spec_to_res_no(residue_spec),
                              coot_utils.residue_spec_to_ins_code(residue_spec))
    for atom_info in ri:
        atom_name = coot_utils.residue_atom2atom_name(atom_info)
        alt_conf  = coot_utils.residue_atom2alt_conf(atom_info)
        coot.add_extra_start_pos_restraint(imol,
                                      res_spec_utils.residue_spec_to_chain_id(residue_spec),
                                      res_spec_utils.residue_spec_to_res_no(residue_spec),
                                      coot_utils.residue_spec_to_ins_code(residue_spec),
                                      atom_name, alt_conf, esd)

        
        pass # fill me



# Copyright 2010, 2011, 2102 Kevin Keating
# 
# Licensed under the Educational Community License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
# 
# http://www.osedu.org/licenses/ECL-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

import inspect, sys, os.path, gtk
import imp

#first, find the location of this script (i.e. the launch.py file) so we know where the RCrane package is located
#note that we can't use __file__ because of the way Coot calls the script, so we use the inspect module instead

fileloc = inspect.currentframe().f_code.co_filename
fileloc = os.path.abspath(fileloc)
rcranePath = os.path.dirname(fileloc)

#import the preLaunchCheck module so we can make sure Coot is able to run RCrane
#we use imp to do the importing in case the RCrane directory name has a period in it
preLaunchCheckPath = os.path.join(rcranePath, "preLaunchCheck.py")
preLaunchCheckFileObject = open(preLaunchCheckPath, "r")
try:
    preLaunchCheck = imp.load_module("rcranePreLaunchCheck", preLaunchCheckFileObject, preLaunchCheckPath, ('.py', 'r', imp.PY_SOURCE))
finally:
    #if the import fails, we still want to close the file object
    if preLaunchCheckFileObject: preLaunchCheckFileObject.close()

#if Coot is able to run RCrane, then load RCrane and create the menu
if preLaunchCheck.checkCootAndReportErrors():
    print "Launching RCrane"
    rcrane = imp.load_module("rcrane", None, rcranePath, ('', '', imp.PKG_DIRECTORY))
        #we use imp to import the rcrane module in case the directory name has a period in it
        #this also allows us to re-import the module later without knowing the directory name
    rcrane.createRCraneMenu()

#destroy any variables that we've created, since the variables will be visible from the Coot Python scripting window
del fileloc
del rcranePath
del preLaunchCheckPath
del preLaunchCheckFileObject
del preLaunchCheck




def hole_ify():
    global start_pos, end_pos
    start_pos = False
    end_pos = False

    def status_bar_pos(position, pos_type):
        s = "Hole %s point set: (%6.2f %6.2f %6.2f)" %(pos_type,
                                                       position[0],
                                                       position[1],
                                                       position[2])
        coot.add_status_bar_text(s)

    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    vbox = gtk.VBox(False, 0)
    hbox = gtk.VBox(False, 0)
    hbox_pos_buttons = gtk.HBox(False, 0)
    hbox_calc_cancel_buttons = gtk.HBox(False, 0)
    start_button = gtk.Button(label="  Set Start Point  ")
    end_button   = gtk.Button(label="  Set End Point  ")
    h_sep = gtk.HSeparator()
    calculate_button = gtk.Button(label="Calculate")
    cancel_button = gtk.Button(label="Cancel")
    hole_export_entry = gtk.Entry()
    export_text = gtk.Label(label="Export surface dots to File: ")
    export_hbox = gtk.HBox(False, 0)
    combobox = coot_gui.generic_molecule_chooser(hbox, "HOLE-ify molecule: ")

    window.add(vbox)
    hbox_pos_buttons.append(start_button)
    hbox_pos_buttons.append(end_button)
    hbox_calc_cancel_buttons.append(calculate_button)
    hbox_calc_cancel_buttons.append(cancel_button)
    export_hbox.append(export_text)
    export_hbox.append(hole_export_entry)
    vbox.append(hbox)
    vbox.append(hbox_pos_buttons)
    vbox.append(export_hbox)
    vbox.pack_start(h_sep)
    vbox.append(hbox_calc_cancel_buttons)

    hole_export_entry.set_text("hole_surface_dots.dat")

    def start_button_cb(*args):
        global start_pos
        start_pos = coot_utils.rotation_centre()
        print("Start pos set to:", start_pos)
        status_bar_pos(start_pos, "start")
        
    start_button.connect("clicked", start_button_cb)

    def end_button_cb(*args):
        global end_pos
        end_pos = coot_utils.rotation_centre()
        print("End pos set to:", end_pos)
        status_bar_pos(end_pos, "end")
        
    end_button.connect("clicked", end_button_cb)

    def delete_event(*args):
        window.destroy()
        return False
    
    def calculate_button_cb(*args):
        global start_pos, end_pos
        imol = combobox_to_molecule_number(combobox)
        if isinstance(imol, int):
            print(start_pos, end_pos)
            if not isinstance(start_pos, list):
                coot.add_status_bar_text("Start position not set")
            else:
                if not isinstance(end_pos, list):
                    coot.add_status_bar_text("End position not set")
                else:
                    # main?
                    print("hole", imol, start_pos, end_pos)
                    # get this from an entry ideally.
                    export_dots_file_name = hole_export_entry.get_text()
                    colour_map_multiplier = 1
                    colour_map_offset = 0
                    hole_args = [imol] + start_pos + end_pos + \
                                [colour_map_multiplier, colour_map_offset, 1,
                                 True, export_dots_file_name]
                    print("BL DEBUG:: hole_args", hole_args)
                    coot.hole(*hole_args)
                    delete_event()
                    
    calculate_button.connect("clicked", calculate_button_cb)

    cancel_button.connect("clicked", delete_event)

    window.show_all()




# group_settings.py
# Copyright 2004, 2006 by Paul Emsley, The University of York
# Copyright 2007 by Bernhard Lohkamp, The University of York
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 

# Add here settings for your group - rather than expecting each user
# to add them to their own .coot files.
# 

# York setting for molprobity. 
# 
#probe_command = "/y/people/lohkamp/coot/Linux/bin/probe.2.12.070727.linuxi386"
#reduce_command = "/y/people/lohkamp/coot/Linux/bin/reduce.3.10.070814.linuxi386"
# try to get the abspath of probe and reduce automatically
# alternatively, specify yourself here.
global probe_command
global reduce_command

probe_command = find_exe("probe", "PATH", "CBIN", "CCP4_BIN")
if not probe_command:
    probe_command = "probe" # useless (?) fallback

reduce_command = find_exe("reduce", "PATH", "CBIN", "CCP4_BIN")
if not reduce_command:
    reduce_command = "reduce" # useless (?) fallback

    



# 08_utils.py
# Copyright 2008 by The University of York
# Author: Bernhard Lohkamp
# Copyright 2007, 2008 by The University of Oxford
# Author: Paul Emsley

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA

import unittest
import os
import coot
import coot_utils
import coot_testing_utils


class TestUtilsFunctions(unittest.TestCase):

    def test01_0(self):
        """Test key symbols"""

        # coot.add_key_binding_gtk3_py("name", "missing", "") # old
        # coot.add_key_binding_gtk3_py(key-code, Ctrl-mode, function, label)

        test_list = [[coot.key_sym_code_py("a-symbol"), -1],
                     [coot.key_sym_code_py(["a", "b"]), -1],
                     [coot.key_sym_code_py("A"), 65],
                     [coot.key_sym_code_py("a"), 97],
                     [coot.key_sym_code_py("9"), 57],
                     [coot.key_sym_code_py("cent"), 162],
                     [coot.key_sym_code_py(">"), 62],
                     [coot.key_sym_code_py(";"), 59],
                     [coot.key_sym_code_py("|"), 124],
                     ]

        for code, result in test_list:
            self.assertEqual(code, result, " fail on key_sym_code, %s is not equal to %s " %(code, result))

    # 20230213-PE No scheme today sadly
    # def test02_0(self):
    #     """Test running a scheme function"""

    #     tot = coot.run_scheme_command("(+ 2 4)")
    #     self.assertEqual(tot, 6)

    #     #run_scheme_command("define test-val 4")
    #     #run_scheme_command("(set! test-val 2)")
    #     rv = coot.run_scheme_command("(rotation-centre)")
    #     print("BL DEBUG:: return scheme is ", rv)
    #     #rv = run_scheme_command("test-val")
    #     #print "BL DEBUG:: return scheme is ", rv
    #     rv = coot.run_scheme_command("2")
    #     self.assertEqual(rv, 2)


    def test03_0(self):
        """Internal/External Molecule Numbers match"""

        m = coot_utils.molecule_number_list()
        #print "   m: ", m
        #print " own: ", map(own_molecule_number, m)
        self.assertTrue(m == list(map(coot.own_molecule_number, m)))


        def test04_0(self):
            """spacegroup operators to space group conversion"""

            test_groups = ["P 1", "P -1", "P 1 2 1", "P 1 1 2", "P 2 1 1", "P 1 21 1", "P 1 1 21",
                           "P 21 1 1", "C 1 2 1", "A 1 2 1", "I 1 2 1", "A 1 1 2", "B 1 1 2", "I 1 1 2",
                           "B 2 1 1", "C 2 1 1", "I 2 1 1", "P 1 m 1", "P 1 1 m", "P m 1 1", "P 1 c 1",
                           "P 1 n 1", "P 1 a 1", "P 1 1 a", "P 1 1 n", "P 1 1 b", "P b 1 1", "P n 1 1",
                           "P c 1 1", "C 1 m 1", "A 1 m 1", "I 1 m 1", "A 1 1 m", "B 1 1 m", "I 1 1 m",
                           "B m 1 1", "C m 1 1", "I m 1 1", "C 1 c 1", "A 1 n 1", "I 1 a 1", "A 1 a 1",
                           "C 1 n 1", "I 1 c 1", "A 1 1 a", "B 1 1 n", "I 1 1 b", "B 1 1 b", "A 1 1 n", "I 1 1 a",
                           "B b 1 1", "C n 1 1", "I c 1 1", "C c 1 1", "B n 1 1", "I b 1 1", "P 1 2/m 1", "P 1 1 2/m",
                           "P 2/m 1 1", "P 1 21/m 1", "P 1 1 21/m", "P 21/m 1 1", "C 1 2/m 1", "A 1 2/m 1", "I 1 2/m 1",
                           "A 1 1 2/m", "B 1 1 2/m", "I 1 1 2/m", "B 2/m 1 1", "C 2/m 1 1", "I 2/m 1 1", "P 1 2/c 1",
                           "P 1 2/n 1", "P 1 2/a 1", "P 1 1 2/a", "P 1 1 2/n", "P 1 1 2/b", "P 2/b 1 1", "P 2/n 1 1",
                           "P 2/c 1 1", "P 1 21/c 1", "P 1 21/n 1", "P 1 21/a 1", "P 1 1 21/a", "P 1 1 21/n", "P 1 1 21/b",
                           "P 21/b 1 1", "P 21/n 1 1", "P 21/c 1 1", "C 1 2/c 1", "A 1 2/n 1", "I 1 2/a 1", "A 1 2/a 1",
                           "C 1 2/n 1", "I 1 2/c 1", "A 1 1 2/a", "B 1 1 2/n", "I 1 1 2/b", "B 1 1 2/b", "A 1 1 2/n",
                           "I 1 1 2/a", "B 2/b 1 1", "C 2/n 1 1", "I 2/c 1 1", "C 2/c 1 1", "B 2/n 1 1", "I 2/b 1 1",
                           "P 2 2 2", "P 2 2 21", "P 21 2 2", "P 2 21 2", "P 21 21 2", "P 2 21 21", "P 21 2 21",
                           "P 21 21 21", "C 2 2 21", "A 21 2 2", "B 2 21 2", "C 2 2 2", "A 2 2 2", "B 2 2 2",
                           "F 2 2 2", "I 2 2 2", "I 21 21 21", "P m m 2", "P 2 m m", "P m 2 m", "P m c 21", "P c m 21",
                           "P 21 m a", "P 21 a m", "P b 21 m", "P m 21 b", "P c c 2", "P 2 a a", "P b 2 b", "P m a 2",
                           "P b m 2", "P 2 m b", "P 2 c m", "P c 2 m", "P m 2 a", "P c a 21", "P b c 21", "P 21 a b",
                           "P 21 c a", "P c 21 b", "P b 21 a", "P n c 2", "P c n 2", "P 2 n a", "P 2 a n", "P b 2 n",
                           "P n 2 b", "P m n 21", "P n m 21", "P 21 m n", "P 21 n m", "P n 21 m", "P m 21 n",
                           "P b a 2", "P 2 c b", "P c 2 a", "P n a 21", "P b n 21", "P 21 n b", "P 21 c n",
                           "P c 21 n", "P n 21 a", "P n n 2", "P 2 n n", "P n 2 n", "C m m 2", "A 2 m m", "B m 2 m",
                           "C m c 21", "C c m 21", "A 21 m a", "A 21 a m", "B b 21 m", "B m 21 b", "C c c 2", "A 2 a a",
                           "B b 2 b", "A m m 2", "B m m 2", "B 2 m m", "C 2 m m", "C m 2 m", "A m 2 m", "A b m 2",
                           "B m a 2", "B 2 c m", "C 2 m b", "C m 2 a", "A c 2 m", "A m a 2", "B b m 2", "B 2 m b", "C 2 c m",
                           "C c 2 m", "A m 2 a", "A b a 2", "B b a 2", "B 2 c b", "C 2 c b", "C c 2 a", "A c 2 a", "F m m 2",
                           "F 2 m m", "F m 2 m", "F d d 2", "F 2 d d", "F d 2 d", "I m m 2", "I 2 m m", "I m 2 m", "I b a 2",
                           "I 2 c b", "I c 2 a", "I m a 2", "I b m 2", "I 2 m b", "I 2 c m", "I c 2 m", "I m 2 a",
                           "P m m m"   "P c c m", "P m a a", "P b m b",
                           "P m m a", "P m m b",
                           "P b m m", "P c m m", "P m c m", "P m a m", "P n n a", "P n n b", "P b n n", "P c n n",
                           "P n c n", "P n a n", "P m n a", "P n m b", "P b m n", "P c n m", "P n c m", "P m a n",
                           "P c c a", "P c c b", "P b a a", "P c a a", "P b c b", "P b a b", "P b a m", "P m c b",
                           "P c m a", "P c c n", "P n a a", "P b n b", "P b c m", "P c a m", "P m c a", "P m a b",
                           "P b m a", "P c m b", "P n n m", "P m n n", "P n m n",
                           "P b c n", "P c a n", "P n c a", "P n a b", "P b n a",
                           "P c n b", "P b c a", "P c a b", "P n m a", "P m n b", "P b n m", "P c m n", "P m c n",
                           "P n a m", "C m c m", "C c m m", "A m m a", "A m a m", "B b m m", "B m m b", "C m c a",
                           "C c m b", "A b m a", "A c a m", "B b c m", "B m a b", "C m m m", "A m m m", "B m m m",
                           "C c c m", "A m a a", "B b m b", "C m m a", "C m m b", "A b m m", "A c m m", "B m c m",
                           "B m a m",
                           "F m m m"  "I m m m", "I b a m", "I m c b",
                           "I c m a", "I b c a", "I c a b", "I m m a", "I m m b", "I b m m", "I c m m", "I m c m",
                           "I m a m", "P 4", "P 41", "P 42", "P 43", "I 4", "I 41", "P -4", "I -4", "P 4/m", "P 42/m",
                           "I 4/m",
                           "P 4 2 2", "P 4 21 2", "P 41 2 2", "P 41 21 2", "P 42 2 2", "P 42 21 2", "P 43 2 2",
                           "P 43 21 2", "I 4 2 2", "I 41 2 2", "P 4 m m", "P 4 b m", "P 42 c m", "P 42 n m",
                           "P 4 c c", "P 4 n c", "P 42 m c", "P 42 b c", "I 4 m m", "I 4 c m", "I 41 m d", "I 41 c d",
                           "P -4 2 m", "P -4 2 c", "P -4 21 m", "P -4 21 c", "P -4 m 2", "P -4 c 2", "P -4 b 2",
                           "P -4 n 2", "I -4 m 2", "I -4 c 2", "I -4 2 m", "I -4 2 d", "P 4/m m m", "P 4/m c c",
                           "P 4/m b m", "P 4/m n c",
                           "P 42/m m c",
                           "P 42/m c m",
                           "P 42/m b c", "P 42/m n m",
                           "I 4/m m m", "I 4/m c m",
                           "P 3", "P 31", "P 32", "P -3",
                           "P 3 1 2", "P 3 2 1", "P 31 1 2", "P 31 2 1", "P 32 1 2", "P 32 2 1",
                           "P 3 m 1", "P 3 1 m", "P 3 c 1", "P 3 1 c",
                           "P -3 1 m", "P -3 1 c", "P -3 m 1", "P -3 c 1",
                           "P 6", "P 61", "P 65", "P 62", "P 64", "P 63", "P -6", "P 6/m",
                           "P 63/m", "P 6 2 2", "P 61 2 2", "P 65 2 2", "P 62 2 2", "P 64 2 2", "P 63 2 2",
                           "P 6 m m", "P 6 c c", "P 63 c m", "P 63 m c", "P -6 m 2", "P -6 c 2", "P -6 2 m",
                           "P -6 2 c", "P 6/m m m", "P 6/m c c", "P 63/m c m", "P 63/m m c", "P 2 3", "F 2 3",
                           "I 2 3", "P 21 3", "I 21 3", "P m -3", "F m -3",
                           "I m -3", "P a -3", "I a -3", "P 4 3 2", "P 42 3 2", "F 4 3 2", "F 41 3 2",
                           "I 4 3 2", "P 43 3 2", "P 41 3 2", "I 41 3 2", "P -4 3 m", "F -4 3 m", "I -4 3 m", "P -4 3 n",
                           "F -4 3 c", "I -4 3 d", "P m -3 m", "P m -3 n",
                           "F m -3 m", "F m -3 c",
                           "I m -3 m", "I a -3 d"]

            # failures: "P n n n :1" "P n n n :2"  "P b a n :1"  "P b a n :2" 
            # "P n c b :1" "P n c b :2" "P c n a :1" "P c n a :2"  "P m m n :1"
            # "P m m n :2"  "P n m m :1"  "P n m m :2" "P m n m :1" "P m n m :2" 
            # "R 3 :H"  "R 3 :R" "R -3 :H" "R -3 :R" "R 3 2 :H"  "R 3 2 :R" 
            # 
            # probable failures (ran out of patience to test the indiviually)
            # "C c c a :1" "C c c a :2" "C c c b :1" "C c c b :2" "A b a a :1"
            # "A b a a :2" "A c a a :1" "A c a a :2" "B b c b :1" "B b c b :2" "B b a b :1"
            # "B b a b :2" "F d d d :1" "F d d d :2"
            # "P 4/n :1" "P 4/n :2" "P 42/n :1" "P 42/n :2"
            # "I 41/a :1" "I 41/a :2" 
            # "P 4/n b m :1" "P 4/n b m :2" "P 4/n n c :1" "P 4/n n c :2" 
            #  "P 4/n m m :1" "P 4/n m m :2" "P 4/n c c :1" "P 4/n c c :2" 
            #  "P 42/n b c :1" "P 42/n b c :2" "P 42/n n m :1" "P 42/n n m :2" 
            # "P 42/n m c :1" "P 42/n m c :2" "P 42/n c m :1" "P 42/n c m :2" 
            # "I 41/a m d :1" "I 41/a m d :2" "I 41/a c d :1" "I 41/a c d :2"
            #  "P n -3 :1" "P n -3 :2"
            # "F d -3 :1"
            # "F d -3 m :1" "F d -3 m :2" "F d -3 c :1"
            # "P n -3 m :1" "F d -3 :2"
            # "R 3 m :H" "R 3 m :R" "R 3 c :H"  "R 3 c :R"  "R -3 m :H" "R -3 m :R"
            #  "R -3 c :H" "R -3 c :R" "P n -3 n :1" "P n -3 n :2" 
            # "P n -3 m :2"
            # "F d -3 c :2" 

            # clipper doesn't know about these space groups (mmdb returns
            # the symop strings correctly (below called symops) (in that
            # they correspond to syminfo.lib):
            #
            # "I 1 21 1" "C 1 21 1"  "B 1 1 m"

            from types import ListType
            from types import StringType

            imol = coot_testing_utils.unittest_pdb("monomer-ACT.pdb")

            for space_group in test_groups:

                set_success = coot.set_space_group(imol, space_group)
                symops = coot.symmetry_operators(imol)

                self.assertTrue(set_success == 1,
                                "   bad status on setting space group %s" %space_group)

                self.assertTrue(type(symops) is ListType,
                                "   bad symops for %s: %s" %(space_group, symops))

                derived_HM = coot.symmetry_operators_to_xHM(symops)

                self.assertTrue(type(derived_HM) is StringType,
                                "   bad derived HM %s" %derived_HM)

                self.assertTrue(space_group == derived_HM,
                                "   No match %s and %s" %(space_group, derived_HM))




# gui_prosmart.py
# Copyright 2007 by Paul Emsley
# Copyright 2007 by The University of Oxford
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import gi
gi.require_version("Gtk", "4.0")
from gi.repository import Gtk, GObject
from gi.repository import Gio
from gi.repository import GLib
import coot
import coot_gui_api
import coot_utils
import coot_gui
from coot_gui import add_simple_action_to_menu, attach_module_menu_button

# target is my molecule, ref is the homologous (high-res) model
#
# extra arg: include_side_chains=False
#
def run_prosmart(imol_target, imol_ref, include_side_chains=False):
    """
    target is my molecule, ref is the homologous (high-res) model

    extra arg: include_side_chains=False
    """

    dir_stub = "coot-ccp4"
    coot.make_directory_maybe(dir_stub)
    r_t_name = coot.molecule_name_stub_py(imol_target, 0).replace(" ", "_");
    r_r_name = coot.molecule_name_stub_py(imol_ref,    0).replace(" ", "_");
    target_pdb_file_name =    os.path.join(dir_stub, r_t_name + "-prosmart.pdb")
    reference_pdb_file_name = os.path.join(dir_stub, r_r_name + "-prosmart-ref.pdb")
    prosmart_out = os.path.join("ProSMART_Output", r_t_name + "-prosmart.txt")

    coot.write_pdb_file(imol_target, target_pdb_file_name)
    coot.write_pdb_file(imol_ref, reference_pdb_file_name)
    prosmart_exe = coot_utils.find_exe("prosmart")
    if prosmart_exe:
        l = ["-p1", target_pdb_file_name,
             "-p2", reference_pdb_file_name,
             "-restrain_seqid", "30"]
        if include_side_chains:
            l += ["-side"]
        coot_utils.popen_command(prosmart_exe,
                                 l,
                                 [],
                                 os.path.join(dir_stub, "prosmart.log"),
                                 False)
        if (not os.path.isfile(prosmart_out)):
            print("file not found", prosmart_out)
        else:
            print("Reading ProSMART restraints from", prosmart_out)
            coot.add_refmac_extra_restraints(imol_target, prosmart_out)
    else:
        coot.info_dialog("No prosmart")


# this should not be in gui_prosmart. Put it in restraints_gui.py
#
def add_module_restraints():

    menu = attach_module_menu_button("Restraints")

    def generate_all_molecule_self_restraints(val):
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no, aa_ins_code, aa_atom_name, aa_alt_conf]:
            coot.generate_self_restraints(aa_imol, val)

    def generate_self_restraint_func(sig):
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no, aa_ins_code, aa_atom_name, aa_alt_conf]:
            print("in generate_self_restraint_func()", aa_imol, aa_chain_id, sig)
            coot.generate_local_self_restraints(aa_imol, aa_chain_id, sig)

    # Generate self restraints for residues around sphere
    def generate_self_restraint_in_sphere_func():
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no, aa_ins_code, aa_atom_name, aa_alt_conf]:
            centred_residue = [aa_chain_id, aa_res_no, aa_ins_code]
            radius = 10
            local_dist_max = 4.2
            other_residues = coot.residues_near_residue_py(aa_imol, centred_residue, radius)
            residue_specs = (centred_residue + other_residues) if isinstance(other_residues, list) else centred_residue
            coot.generate_local_self_restraints_by_residues_py(aa_imol, residue_specs, local_dist_max)

    def display_extra_restraints_func(state):
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                    aa_ins_code, aa_atom_name, aa_alt_conf]:
            # this is a bit confusing
            coot.set_show_extra_restraints(aa_imol, state)
            coot.set_show_extra_distance_restraints(state)

    def delete_all_extra_restraints_func():
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                              aa_ins_code, aa_atom_name, aa_alt_conf]:
            coot.delete_all_extra_restraints(aa_imol)

    add_simple_action_to_menu(
        menu, "Generate Self Restraints 3.7 for Chain", "generate_self_restraint_37",
        lambda _simple_action, _arg: generate_self_restraint_func(3.7))

    add_simple_action_to_menu(
        menu, "Generate Chain Self Restraints 4.3 for Chain", "generate_self_restraint_43",
        lambda _simple_action, _arg: generate_self_restraint_func(4.3))

    add_simple_action_to_menu(
        menu, "Generate Chain Self Restraints 5.0 for Chain", "generate_self_restraint_50",
        lambda _simple_action, _arg: generate_self_restraint_func(5.0))

    add_simple_action_to_menu(
        menu, "Generate Chain Self Restraints 6.0 for Chain", "generate_self_restraint_60",
        lambda _simple_action, _arg: generate_self_restraint_func(6))

    add_simple_action_to_menu(
        menu, "Generate Local Self Restraints 6", "generate_self_restraint_in_sphere",
        lambda _simple_action, _arg: generate_self_restraint_in_sphere_func())

    add_simple_action_to_menu(menu, "Generate All-molecule Self Restraints 4.3",
                              "generate_all_molecule_self_restraints_43",
                              lambda _simple_action, _arg: generate_all_molecule_self_restraints(4.3))

    add_simple_action_to_menu(menu, "Generate All-molecule Self Restraints 5.0",
                              "generate_all_molecule_self_restraints_50",
                              lambda _simple_action, _arg: generate_all_molecule_self_restraints(5.0))

    add_simple_action_to_menu(menu, "Generate All-molecule Self Restraints 6.0",
                              "generate_all_molecule_self_restraints_60",
                              lambda _simple_action, _arg: generate_all_molecule_self_restraints(6.0))

    add_simple_action_to_menu(menu, "Delete All Extra Restraints", "delete_all_extra_restraints",
                              lambda _simple_action, _arg: delete_all_extra_restraints_func())


    # todo: Move it under R/RC and make it a single toggle

    # add_simple_action_to_menu(
    #     menu, "Undisplay Extra Restraints","undisplay_extra_restraints",
    #     lambda _simple_action, _arg: display_extra_restraints_func(0))

    # add_simple_action_to_menu(
    #     menu, "Display Extra Restraints","display_extra_restraints",
    #     lambda _simple_action, _arg: display_extra_restraints_func(1))


def add_module_prosmart():

    menu = attach_module_menu_button("ProSMART")

    def combobox_to_molecule_number(combobox):
        imol = -1
        tree_iter = combobox.get_active_iter()
        if tree_iter is not None:
            model = combobox.get_model()
            it = model[tree_iter]
            imol = it[0]
        return imol

    def launch_prosmart_gui():
        def go_button_cb(*args):
            imol_tar = combobox_to_molecule_number(combobox_tar)
            imol_ref = combobox_to_molecule_number(combobox_ref)
            do_side_chains = check_button.get_active()
            run_prosmart(imol_tar, imol_ref, do_side_chains)
            window.destroy()
        window = Gtk.Window()
        window.set_size_request(240,180)
        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL,spacing=5)
        hbox.set_homogeneous(True)
        hbox.set_halign(Gtk.Align.END)
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL,spacing=10)
        vbox.set_margin_start(10)
        vbox.set_margin_end(10)
        vbox.set_margin_top(10)
        vbox.set_margin_bottom(10)
        h_sep = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
        chooser_hint_text_1 = "Target molecule"
        chooser_hint_text_2 = "Reference (high-res) molecule"
        go_button = Gtk.Button(label="ProSMART")
        cancel_button = Gtk.Button(label="Cancel")
        check_button = Gtk.CheckButton(label="Include Side-chains")

        combobox_tar = coot_gui.generic_molecule_chooser(vbox, chooser_hint_text_1)
        combobox_ref = coot_gui.generic_molecule_chooser(vbox, chooser_hint_text_2)

        vbox.append(check_button)
        vbox.append(h_sep)
        vbox.append(hbox)
        hbox.append(cancel_button)
        hbox.append(go_button)
        window.set_child(vbox)

        cancel_button.connect("clicked", lambda _w: window.close())

        go_button.connect("clicked", go_button_cb)
        window.show()

    def generate_self_restraint_func(sig):
        # is this used?
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no, aa_ins_code, aa_atom_name, aa_alt_conf]:
            generate_local_self_restraints(aa_imol, aa_chain_id, sig)

    def prosmart_cut_to_func(sig_low, sig_high):
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no, aa_ins_code, aa_atom_name, aa_alt_conf]:
            coot.set_extra_restraints_prosmart_sigma_limits(aa_imol, sig_low, sig_high)
    add_simple_action_to_menu(
        menu, "Show Only Deviant Distances Beyond 6","prosmart_cut_to_6",
        lambda _simple_action, _arg: prosmart_cut_to_func(-6, 6))

    add_simple_action_to_menu(
        menu, "Show Only Deviant Distances Beyond 4","prosmart_cut_to_4",
        lambda _simple_action, _arg: prosmart_cut_to_func(-4, 4))

    add_simple_action_to_menu(
        menu, "Show Only Deviant Distances Beyond 2.0","prosmart_cut_to_2",
        lambda _simple_action, _arg: prosmart_cut_to_func(-2, 2))

    add_simple_action_to_menu(
        menu, "Show Only Deviant Distances Beyond 1.0","prosmart_cut_to_1",
        lambda _simple_action, _arg: prosmart_cut_to_func(-1, 1))

    add_simple_action_to_menu(
        menu, "Undisplay All Extra Distance Restraints","prosmart_cut_to_0",
        lambda _simple_action, _arg: prosmart_cut_to_func(0, 0))

    def restraint_to_ca_func(state):
        with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
                                    aa_ins_code, aa_atom_name, aa_alt_conf]:
            coot.set_extra_restraints_representation_for_bonds_go_to_CA(aa_imol, state)

    # # clutter
    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Restraint Representation To CA",
    #     lambda func: restraint_to_ca_func(1))

    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Restraint Representation To Home Atom",
    #     lambda func: restraint_to_ca_func(0))

    add_simple_action_to_menu(
        menu, "Run ProSMART...","run_prosmart_gui", lambda _simple_action, _arg: launch_prosmart_gui())

    ## extra
    
    # def delete_all_extra_restraints_func():
    #     with coot_utils.UsingActiveAtom() as [aa_imol, aa_chain_id, aa_res_no,
    #                                aa_ins_code, aa_atom_name, aa_alt_conf]:
    #         coot.delete_all_extra_restraints(aa_imol)

    # coot_gui.add_simple_coot_menu_menuitem(
    #     menu, "Delete All Extra Restraints",
    #     lambda func: delete_all_extra_restraints_func())

